from dataclasses import dataclass, field
from ..substrate import linestringToKicad
from ..defs import Layer
from ..common import KiAngle, KiLength, fromDegrees, fromMm
from ..pcbnew_utils import increaseZonePriorities
import pcbnew
from ..panelize import Panel
from .baseFeature import PanelFeature
from typing import Any, List, Tuple
import numpy as np
from shapely.geometry import (
    Polygon,
    MultiPolygon)
from shapely.ops import unary_union

class KiCADCopperFillMixin(PanelFeature):
    """
    Build solid infill of non-board areas
    """
    def _adjustZoneParameters(self, zone: pcbnew.ZONE):
        """
        Allow an inherited class to override KiCAD zone parameters
        """
        pass  # solid infill does nothing

    def apply(self, panel: Any) -> None:
        if not len(self.layers) > 0:
            raise RuntimeError("No layers to add copper to")
        increaseZonePriorities(panel.board)

        zoneArea = panel.boardSubstrate.substrates.buffer(-self.edgeclearance)
        zoneArea = zoneArea.difference(unary_union(
            [substrate.exterior().buffer(self.clearance) for substrate in panel.substrates]
        ))

        geoms = [zoneArea] if isinstance(zoneArea, Polygon) else zoneArea.geoms

        for g in geoms:
            if len(g.exterior.coords) == 0:
                # Skip empty geometries
                continue
            zoneContainer = pcbnew.ZONE(panel.board)
            self._adjustZoneParameters(zoneContainer)
            zoneContainer.Outline().AddOutline(linestringToKicad(g.exterior))
            for hole in g.interiors:
                zoneContainer.Outline().AddHole(linestringToKicad(hole))
            zoneContainer.SetAssignedPriority(0)

            for l in self.layers:
                panel._ensureLayerEnabled(l)
                zoneContainer = zoneContainer.Duplicate()
                zoneContainer.SetLayer(l)
                panel.board.Add(zoneContainer)
                panel.zonesToRefill.append(zoneContainer)


@dataclass
class SolidCopperFill(KiCADCopperFillMixin):
    """
    Build solid infill of non-board areas
    """
    clearance: KiLength = field(default_factory=lambda: fromMm(1))
    edgeclearance: KiLength = field(default_factory=lambda: fromMm(1))
    layers: List[Layer] = field(default_factory=lambda: [Layer.F_Cu, Layer.B_Cu])

    def _adjustZoneParameters(self, zone: pcbnew.ZONE) -> None:
        pass # There are no adjustments for solid infill


@dataclass
class HatchedCopperFill(KiCADCopperFillMixin):
    """
    Build hatched infill of non-board areas
    """
    clearance: KiLength = field(default_factory=lambda: fromMm(1))
    edgeclearance: KiLength = field(default_factory=lambda: fromMm(1))
    layers: List[Layer] = field(default_factory=lambda: [Layer.F_Cu, Layer.B_Cu])
    strokeWidth: KiLength = field(default_factory=lambda: fromMm(1))
    strokeSpacing: KiLength = field(default_factory=lambda: fromMm(1))
    orientation: KiAngle = field(default_factory=lambda: fromDegrees(45))

    def _adjustZoneParameters(self, zoneContainer: pcbnew.ZONE) -> None:
        zoneContainer.SetFillMode(pcbnew.ZONE_FILL_MODE_HATCH_PATTERN)
        zoneContainer.SetHatchOrientation(self.orientation)
        zoneContainer.SetHatchGap(self.strokeSpacing)
        zoneContainer.SetHatchThickness(self.strokeWidth)

@dataclass
class HexCopperFill(PanelFeature):
    """
    Build hex infill of non-board areas
    """
    clearance: KiLength = field(default_factory=lambda: fromMm(1))
    edgeclearance: KiLength = field(default_factory=lambda: fromMm(1))
    layers: List[Layer] = field(default_factory=lambda: [Layer.F_Cu, Layer.B_Cu])
    diameter: KiLength = field(default_factory=lambda: fromMm(7))
    space: KiLength = field(default_factory=lambda: fromMm(0.5))
    threshold: float = field(default_factory=lambda: 0.25)

    def _buildHexagonsPolygon(self, area: Tuple[float, float, float, float]) -> MultiPolygon:
        horizontalSpacing = self.space + np.sqrt(3) / 2 * self.diameter
        verticalSpacing = 3 / 4 * self.diameter + np.sqrt(3) / 2 * self.space

        minx, miny, maxx, maxy = area

        maxx += horizontalSpacing
        maxy += horizontalSpacing

        hexagons = []
        y = miny
        shifted = False
        while y <= maxy:
            x = minx - (horizontalSpacing / 2 if shifted else 0)
            while x <= maxx:
                hexagons.append(Polygon([
                    (x + self.diameter / 2 * np.cos(np.pi / 6 + i / 3 * np.pi),
                     y + self.diameter / 2 * np.sin(np.pi / 6 + i / 3 * np.pi)) for i in range(6)
                ]))
                x += horizontalSpacing
            y += verticalSpacing
            shifted = not shifted

        return MultiPolygon(hexagons)

    def apply(self, panel: Panel) -> None:
        if not len(self.layers) > 0:
            raise RuntimeError("No layers to add copper to")

        increaseZonePriorities(panel.board)

        zoneArea = panel.boardSubstrate.substrates.buffer(-self.edgeclearance)
        zoneArea = zoneArea.intersection(panel.boardSubstrate.substrates)
        zoneArea = zoneArea.difference(unary_union(
            [substrate.exterior().buffer(self.clearance) for substrate in panel.substrates]
        ))

        hexagons = self._buildHexagonsPolygon(zoneArea.bounds)
        hexagons = hexagons.intersection(zoneArea)

        baseHexArea = 3 * np.sqrt(3) * (self.diameter / 2) ** 2 / 2

        geoms = [hexagons] if isinstance(hexagons, Polygon) else hexagons.geoms
        for g in geoms:
            if g.area < self.threshold * baseHexArea:
                continue
            zoneContainer = pcbnew.ZONE(panel.board)
            zoneContainer.Outline().AddOutline(linestringToKicad(g.exterior))
            for hole in g.interiors:
                zoneContainer.Outline().AddHole(linestringToKicad(hole))
            zoneContainer.SetAssignedPriority(0)

            for l in self.layers:
                panel._ensureLayerEnabled(l)
                zoneContainer = zoneContainer.Duplicate()
                zoneContainer.SetLayer(l)
                panel.board.Add(zoneContainer)
                panel.zonesToRefill.append(zoneContainer)



import click
import pcbnew
import csv
import os
import sys
import shutil
from pathlib import Path
from kikit.fab.common import *
from kikit.common import *
from kikit.export import gerberImpl

def collectBom(components, lscsFields, ignore):
    bom = {}
    for c in components:
        if getUnit(c) != 1:
            continue
        reference = getReference(c)
        if reference.startswith("#PWR") or reference.startswith("#FL"):
            continue
        if reference in ignore:
            continue
        if getField(c, "JLCPCB_IGNORE") is not None and getField(c, "JLCPCB_IGNORE") != "":
            continue
        if hasattr(c, "in_bom") and not c.in_bom:
            continue
        if hasattr(c, "on_board") and not c.on_board:
            continue
        if hasattr(c, "dnp") and c.dnp:
            continue
        orderCode = None
        for fieldName in lscsFields:
            orderCode = getField(c, fieldName)
            if orderCode is not None and orderCode.strip() != "":
                break
        cType = (
            getField(c, "Value"),
            getField(c, "Footprint"),
            orderCode
        )
        bom[cType] = bom.get(cType, []) + [reference]
    return bom

def bomToCsv(bomData, filename):
    with open(filename, "w", newline="", encoding="utf-8") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(["Comment", "Designator", "Footprint", "LCSC"])
        for cType, references in bomData.items():
            # JLCPCB allows at most 200 components per line so we have to split
            # the BOM into multiple lines. Let's make the chunks by 100 just to
            # be sure.
            CHUNK_SIZE = 100
            sortedReferences = sorted(references, key=naturalComponentKey)
            for i in range(0, len(references), CHUNK_SIZE):
                refChunk = sortedReferences[i:i+CHUNK_SIZE]
                value, footprint, lcsc = cType
                writer.writerow([value, ",".join(refChunk), footprint, lcsc])

def exportJlcpcb(board, outputdir, assembly, schematic, ignore, field,
           corrections, correctionpatterns, missingerror, nametemplate, drc,
           autoname):
    """
    Prepare fabrication files for JLCPCB including their assembly service
    """
    ensureValidBoard(board)
    loadedBoard = pcbnew.LoadBoard(board)

    if drc:
        ensurePassingDrc(loadedBoard)

    refsToIgnore = parseReferences(ignore)
    removeComponents(loadedBoard, refsToIgnore)
    Path(outputdir).mkdir(parents=True, exist_ok=True)

    gerberdir = os.path.join(outputdir, "gerber")
    shutil.rmtree(gerberdir, ignore_errors=True)
    gerberImpl(board, gerberdir)

    if autoname:
        boardName = os.path.basename(board.replace(".kicad_pcb", ""))
        archiveName = expandNameTemplate(nametemplate, boardName + "-gerbers", loadedBoard)
    else:
        archiveName = expandNameTemplate(nametemplate, "gerbers", loadedBoard)
    shutil.make_archive(os.path.join(outputdir, archiveName), "zip", outputdir, "gerber")

    if not assembly:
        return
    if schematic is None:
        raise RuntimeError("When outputing assembly data, schematic is required")

    ensureValidSch(schematic)

    correctionFields = [x.strip() for x in corrections.split(",")]
    components = extractComponents(schematic)
    ordercodeFields = [x.strip() for x in field.split(",")]
    bom = collectBom(components, ordercodeFields, refsToIgnore)

    bom_refs = set(x for xs in bom.values() for x in xs)
    bom_components = [c for c in components if getReference(c) in bom_refs]

    posData = collectPosData(loadedBoard, correctionFields,
        bom=bom_components, posFilter=noFilter, correctionFile=correctionpatterns,
        orientationHandling=FootprintOrientationHandling.MirrorBottom)
    boardReferences = set([x[0] for x in posData])
    bom = {key: [v for v in val if v in boardReferences] for key, val in bom.items()}
    bom = {key: val for key, val in bom.items() if len(val) > 0}


    missingFields = False
    for type, references in bom.items():
        _, _, lcsc = type
        if not lcsc:
            missingFields = True
            for r in references:
                print(f"WARNING: Component {r} is missing ordercode")
    if missingFields and missingerror:
        sys.exit("There are components with missing ordercode, aborting")

    posDataToFile(posData, os.path.join(outputdir, expandNameTemplate(nametemplate, "pos", loadedBoard) + ".csv"))
    bomToCsv(bom, os.path.join(outputdir, expandNameTemplate(nametemplate, "bom", loadedBoard) + ".csv"))



from ..panelize import Panel

class PanelFeature:
    """
    Basic interface for various
    """
    def apply(self, panel: Panel) -> None:
        raise NotImplementedError("Implementation error: PanelFeature doesn't support applying")




from kikit.pcbnew_utils import KICAD_VERSION
from kikit.common import KIKIT_LIB
import sys
import click

@click.command()
def kicadversion():
    """
    Return version of KiCAD
    """
    print(f"{KICAD_VERSION[0]}.{KICAD_VERSION[1]}")

@click.command()
def drcapi():
    """
    Return version of the DRC API
    """
    print("1")

@click.command()
def lib():
    """
    Return KiKit library location
    """
    print(KIKIT_LIB)


@click.group()
def cli():
    """
    Get information about the KiCAD installation
    """
    pass

cli.add_command(kicadversion)
cli.add_command(drcapi)
cli.add_command(lib)



import pcbnew
import os
import shutil
from pathlib import Path
from kikit.export import gerberImpl, exportSettingsOSHPark, fullGerberPlotPlan
from kikit.fab.common import ensurePassingDrc, expandNameTemplate

plotPlanNoVCuts = [(name, id, comment) for name, id, comment in fullGerberPlotPlan if name != "CmtUser"]

def exportOSHPark(board, outputdir, nametemplate, drc):
    """
    Prepare fabrication files for OSH Park
    """
    loadedBoard = pcbnew.LoadBoard(board)
    Path(outputdir).mkdir(parents=True, exist_ok=True)

    if drc:
        ensurePassingDrc(loadedBoard)

    gerberdir = os.path.join(outputdir, "gerber")
    shutil.rmtree(gerberdir, ignore_errors=True)
    gerberImpl(board, gerberdir, plot_plan=plotPlanNoVCuts, settings=exportSettingsOSHPark)

    archiveName = expandNameTemplate(nametemplate, "gerbers", loadedBoard)
    shutil.make_archive(os.path.join(outputdir, archiveName), "zip", outputdir, "gerber")



import pcbnew
import numpy as np
import json
from collections import OrderedDict
from kikit.common import *
from kikit.defs import *
from kikit.substrate import Substrate, extractRings, toShapely, linestringToKicad
from kikit.export import gerberImpl, pasteDxfExport
from kikit.export import exportSettingsJlcpcb
import solid
import solid.utils
import subprocess
import shutil
from kikit.common import removeComponents, parseReferences

from shapely.geometry import Point


OUTER_BORDER = fromMm(7.5)
INNER_BORDER = fromMm(5)
MOUNTING_HOLES_COUNT = 3
MOUNTING_HOLE_R = fromMm(1)
HOLE_SPACING = fromMm(20)

def addBottomCounterpart(board, item):
    item = item.Duplicate()
    item.SetLayer(Layer.B_Paste)
    board.Add(item)

def addRoundedCorner(board, center, start, end, thickness):
    corner = pcbnew.PCB_SHAPE()
    corner.SetShape(STROKE_T.S_ARC)
    corner.SetCenter(toKiCADPoint((center[0], center[1])))
    corner.SetStart(toKiCADPoint((start[0], start[1])))

    if np.cross(start - center, end - center) > 0:
        corner.SetArcAngleAndEnd(fromDegrees(90), True)
    else:
        corner.SetArcAngleAndEnd(fromDegrees(-90), True)
    corner.SetWidth(thickness)
    corner.SetLayer(Layer.F_Paste)
    board.Add(corner)
    addBottomCounterpart(board, corner)

def addLine(board, start, end, thickness):
    line = pcbnew.PCB_SHAPE()
    line.SetShape(STROKE_T.S_SEGMENT)
    line.SetStart(toKiCADPoint((start[0], start[1])))
    line.SetEnd(toKiCADPoint((end[0], end[1])))
    line.SetWidth(thickness)
    line.SetLayer(Layer.F_Paste)
    board.Add(line)
    addBottomCounterpart(board, line)

def addBite(board, origin, direction, normal, thickness):
    """
    Adds a bite to the stencil, direction points to the bridge, normal points
    inside the stencil
    """
    direction = normalize(direction) * thickness
    normal = normalize(normal) * thickness
    center = toKiCADPoint((origin[0], origin[1])) + toKiCADPoint((normal[0], normal[1]))
    start = origin
    end = center + toKiCADPoint((direction[0], direction[1]))
    # addLine(board, end, end + normal / 2, thickness)
    addRoundedCorner(board, center, start, end, thickness)

def numberOfCuts(length, bridgeWidth, bridgeSpacing):
    """
    Return number of bridges which fit inside the length and cut length
    """
    count = int(np.floor((length + bridgeWidth) / (bridgeWidth + bridgeSpacing)))
    cutLength = (length - (count - 1) * bridgeWidth) / count
    return count, cutLength

def addFrame(board, rect, bridgeWidth, bridgeSpacing, clearance):
    """
    Add rectangular frame to the board
    """
    R=fromMm(1)

    corners = [
        (tl(rect), toKiCADPoint((R, 0)), toKiCADPoint((0, R))), # TL
        (tr(rect), toKiCADPoint((0, R)), toKiCADPoint((-R, 0))), # TR
        (br(rect), toKiCADPoint((-R, 0)), toKiCADPoint((0, -R))), # BR
        (bl(rect), toKiCADPoint((0, -R)), toKiCADPoint((R, 0))) # BL
    ]
    for c, sOffset, eOffset in corners:
        addRoundedCorner(board, c + sOffset + eOffset, c + sOffset, c + eOffset, clearance)

    count, cutLength = numberOfCuts(rect.GetWidth() - 2 * R, bridgeWidth, bridgeSpacing)
    for i in range(count):
        start = rect.GetX() + R + i * bridgeWidth + i * cutLength
        end = start + cutLength

        y1, y2 = rect.GetY(), rect.GetY() + rect.GetHeight()
        addLine(board, toKiCADPoint((start, y1)), toKiCADPoint((end, y1)), clearance)
        if i != 0:
            addBite(board, toKiCADPoint((start, y1)), toKiCADPoint((-1, 0)), toKiCADPoint((0, 1)), clearance)
        if i != count - 1:
            addBite(board, toKiCADPoint((end, y1)), toKiCADPoint((1, 0)), toKiCADPoint((0, 1)), clearance)
        addLine(board, toKiCADPoint((start, y2)), toKiCADPoint((end, y2)), clearance)
        if i != 0:
            addBite(board, toKiCADPoint((start, y2)), toKiCADPoint((-1, 0)), toKiCADPoint((0, -1)), clearance)
        if i != count - 1:
            addBite(board, toKiCADPoint((end, y2)), toKiCADPoint((1, 0)), toKiCADPoint((0, -1)), clearance)

    count, cutLength = numberOfCuts(rect.GetHeight() - 2 * R, bridgeWidth, bridgeSpacing)
    for i in range(count):
        start = rect.GetY() + R + i * bridgeWidth + i * cutLength
        end = start + cutLength

        x1, x2 = rect.GetX(), rect.GetX() + rect.GetWidth()
        addLine(board, toKiCADPoint((x1, start)), toKiCADPoint((x1, end)), clearance)
        if i != 0:
            addBite(board, toKiCADPoint((x1, start)), toKiCADPoint((0, -1)), toKiCADPoint((1, 0)), clearance)
        if i != count - 1:
            addBite(board, toKiCADPoint((x1, end)), toKiCADPoint((0, 1)), toKiCADPoint((1, 0)), clearance)
        addLine(board, toKiCADPoint((x2, start)), toKiCADPoint((x2, end)), clearance)
        if i != 0:
            addBite(board, toKiCADPoint((x2, start)), toKiCADPoint((0, -1)), toKiCADPoint((-1, 0)), clearance)
        if i != count - 1:
            addBite(board, toKiCADPoint((x2, end)), toKiCADPoint((0, 1)), toKiCADPoint((-1, 0)), clearance)

def addHole(board, position, radius):
    circle = pcbnew.PCB_SHAPE()
    circle.SetShape(STROKE_T.S_CIRCLE)
    circle.SetCenter(toKiCADPoint((position[0], position[1])))
    # Set 3'oclock point of the circle to set radius
    circle.SetEnd(toKiCADPoint((position[0], position[1])) + toKiCADPoint((radius/2, 0)))

    circle.SetWidth(radius)
    circle.SetLayer(Layer.F_Paste)
    board.Add(circle)
    addBottomCounterpart(board, circle)

def addJigFrame(board, jigFrameSize, bridgeWidth=fromMm(2),
                bridgeSpacing=fromMm(10), clearance=fromMm(0.5)):
    """
    Given a Pcbnew board finds the board outline and creates a stencil for
    KiKit's stencil jig.

    Mainly, adds mounting holes and mouse bites to define the panel outline.

    jigFrameSize is a tuple (width, height).
    """
    bBox = findBoardBoundingBox(board)
    frameSize = rectByCenter(rectCenter(bBox),
        jigFrameSize[0] + 2 * (OUTER_BORDER + INNER_BORDER),
        jigFrameSize[1] + 2 * (OUTER_BORDER + INNER_BORDER))
    cutSize = rectByCenter(rectCenter(bBox),
        jigFrameSize[0] + 2 * (OUTER_BORDER + INNER_BORDER) - fromMm(1),
        jigFrameSize[1] + 2 * (OUTER_BORDER + INNER_BORDER) - fromMm(1))
    addFrame(board, cutSize, bridgeWidth, bridgeSpacing, clearance)

    for i in range(MOUNTING_HOLES_COUNT):
        x = frameSize.GetX() + OUTER_BORDER / 2 + (i + 1) * (frameSize.GetWidth() - OUTER_BORDER) / (MOUNTING_HOLES_COUNT + 1)
        addHole(board, toKiCADPoint((x, OUTER_BORDER / 2 + frameSize.GetY())), MOUNTING_HOLE_R)
        addHole(board, toKiCADPoint((x, - OUTER_BORDER / 2 +frameSize.GetY() + frameSize.GetHeight())), MOUNTING_HOLE_R)
    for i in range(MOUNTING_HOLES_COUNT):
        y = frameSize.GetY() + OUTER_BORDER / 2 + (i + 1) * (frameSize.GetHeight() - OUTER_BORDER) / (MOUNTING_HOLES_COUNT + 1)
        addHole(board, toKiCADPoint((OUTER_BORDER / 2 + frameSize.GetX(), y)), MOUNTING_HOLE_R)
        addHole(board, toKiCADPoint((- OUTER_BORDER / 2 +frameSize.GetX() + frameSize.GetWidth(), y)), MOUNTING_HOLE_R)

    PIN_TOLERANCE = fromMm(0.05)
    addHole(board, tl(frameSize) + toKiCADPoint((OUTER_BORDER / 2, OUTER_BORDER / 2)), MOUNTING_HOLE_R + PIN_TOLERANCE)
    addHole(board, tr(frameSize) + toKiCADPoint((-OUTER_BORDER / 2, OUTER_BORDER / 2)), MOUNTING_HOLE_R + PIN_TOLERANCE)
    addHole(board, br(frameSize) + toKiCADPoint((-OUTER_BORDER / 2, -OUTER_BORDER / 2)), MOUNTING_HOLE_R + PIN_TOLERANCE)
    addHole(board, bl(frameSize) + toKiCADPoint((OUTER_BORDER / 2, -OUTER_BORDER / 2)), MOUNTING_HOLE_R + PIN_TOLERANCE)

def jigMountingHoles(jigFrameSize, origin=toKiCADPoint((0, 0))):
    """ Get list of all mounting holes in a jig of given size """
    w, h = jigFrameSize
    holes = [
        toKiCADPoint((0, (w + INNER_BORDER) / 2)),
        toKiCADPoint((0, -(w + INNER_BORDER) / 2)),
        toKiCADPoint(((h + INNER_BORDER) / 2, 0)),
        toKiCADPoint((-(h + INNER_BORDER) / 2, 0)),
    ]
    return [x + origin for x in holes]

def createOuterPolygon(board, jigFrameSize, outerBorder):
    bBox = findBoardBoundingBox(board)
    centerpoint = rectCenter(bBox)
    holes = jigMountingHoles(jigFrameSize, centerpoint)

    outerSubstrate = Substrate(collectEdges(board, Layer.Edge_Cuts))
    outerSubstrate.substrates = outerSubstrate.substrates.buffer(outerBorder)
    tabs = []
    for hole in holes:
        tab, _ = outerSubstrate.tab(hole, centerpoint - hole, INNER_BORDER, maxHeight=fromMm(1000))
        tabs.append(tab)
    outerSubstrate.union(tabs)
    outerSubstrate.union([Point(x).buffer(INNER_BORDER / 2) for x in holes])
    outerSubstrate.millFillets(fromMm(3))
    return outerSubstrate.exterior(), holes

def createOffsetPolygon(board, offset):
    outerSubstrate = Substrate(collectEdges(board, Layer.Edge_Cuts))
    outerSubstrate.substrates = outerSubstrate.substrates.buffer(offset)
    return outerSubstrate.exterior()

def m2countersink():
    HEAD_DIA = fromMm(4.5)
    HOLE_LEN = fromMm(10)
    SINK_EXTRA = fromMm(0.3)
    sinkH = np.sqrt(HEAD_DIA**2 / 4)

    sink = solid.cylinder(d1=0, d2=HEAD_DIA, h=sinkH)
    sinkE = solid.cylinder(d=HEAD_DIA, h=SINK_EXTRA)
    hole = solid.cylinder(h=HOLE_LEN, d=fromMm(2))
    return sinkE + solid.utils.down(sinkH)(sink) + solid.utils.down(HOLE_LEN)(hole)

def mirrorX(linestring, origin):
    return [(2 * origin - x, y) for x, y in linestring]

def makeRegister(board, jigFrameSize, jigThickness, pcbThickness,
                 outerBorder, innerBorder, tolerance, topSide):
    bBox = findBoardBoundingBox(board)
    centerpoint = rectCenter(bBox)

    top = jigThickness - fromMm(0.15)
    pcbBottom = jigThickness - pcbThickness

    outerPolygon, holes = createOuterPolygon(board, jigFrameSize, outerBorder)
    outerRing = outerPolygon.exterior.coords
    if topSide:
        outerRing = mirrorX(outerRing, centerpoint[0])
    body = solid.linear_extrude(height=top, convexity=10)(solid.polygon(
        outerRing))

    innerRings = [x.exterior.coords for x in listGeometries(createOffsetPolygon(board, - innerBorder))]
    if topSide:
        innerRings = [mirrorX(innerRing, centerpoint[0]) for innerRing in innerRings]

    innerCutout = solid.utils.down(jigThickness)(
        solid.linear_extrude(height=3 * jigThickness, convexity=10)(solid.polygon(innerRings[0])))
    for innerRing in innerRings[1:]:
        innerCutout = innerCutout + solid.utils.down(jigThickness)(
            solid.linear_extrude(height=3 * jigThickness, convexity=10)(solid.polygon(innerRing)))
    registerRing = createOffsetPolygon(board, tolerance).exterior.coords
    if topSide:
        registerRing = mirrorX(registerRing, centerpoint[0])
    registerCutout = solid.utils.up(jigThickness - pcbThickness)(
        solid.linear_extrude(height=jigThickness, convexity=10)(solid.polygon(registerRing)))

    register = body - innerCutout - registerCutout
    for hole in holes:
        register = register - solid.translate([hole[0], hole[1], top])(m2countersink())
    return solid.scale(toMm(1))(
            solid.translate([-centerpoint[0], -centerpoint[1], 0])(register))

def makeTopRegister(board, jigFrameSize, jigThickness, pcbThickness,
                    outerBorder=fromMm(3), innerBorder=fromMm(1),
                    tolerance=fromMm(0.05)):
    """
    Create a SolidPython representation of the top register
    """
    return makeRegister(board, jigFrameSize, jigThickness, pcbThickness,
            outerBorder, innerBorder, tolerance, True)

def makeBottomRegister(board, jigFrameSize, jigThickness, pcbThickness,
                    outerBorder=fromMm(3), innerBorder=fromMm(1),
                    tolerance=fromMm(0.05)):
    """
    Create a SolidPython representation of the top register
    """
    return makeRegister(board, jigFrameSize, jigThickness, pcbThickness,
            outerBorder, innerBorder, tolerance, False)

def renderScad(infile, outfile):
    infile = os.path.abspath(infile)
    outfile = os.path.abspath(outfile)
    try:
        subprocess.run(["openscad", "-o", outfile, infile],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError as e:
        message = f"Cannot render {outfile}, OpenSCAD error:\n"
        message += (e.stdout.decode("utf-8") + "\n") if e.stdout is not None else ""
        message += (e.stderr.decode("utf-8") + "\n") if e.stderr is not None else ""
        raise RuntimeError(message)
    except FileNotFoundError as e:
        message = f"OpenSCAD is not available.\n"
        message += f"Did you install it? Program `openscad` has to be in PATH"
        raise RuntimeError(message)

def shapelyToSHAPE_POLY_SET(polygon):
    p = pcbnew.SHAPE_POLY_SET()
    p.AddOutline(linestringToKicad(polygon.exterior))
    return p

def cutoutComponents(board, components):
    topCutout = extractComponentPolygons(components, pcbnew.F_CrtYd)
    for polygon in topCutout:
        zone = pcbnew.PCB_SHAPE()
        zone.SetShape(STROKE_T.S_POLYGON)
        zone.SetPolyShape(shapelyToSHAPE_POLY_SET(polygon))
        zone.SetLayer(Layer.F_Paste)
        board.Add(zone)
    bottomCutout = extractComponentPolygons(components, pcbnew.B_CrtYd)
    for polygon in bottomCutout:
        zone = pcbnew.PCB_SHAPE()
        zone.SetShape(STROKE_T.S_POLYGON)
        zone.SetPolyShape(shapelyToSHAPE_POLY_SET(polygon))
        zone.SetLayer(Layer.B_Paste)
        board.Add(zone)

def setStencilLayerVisibility(boardName):
    prlPath = os.path.splitext(boardName)[0] + ".kicad_prl"
    try:
        with open(prlPath, encoding="utf-8") as f:
            # We use ordered dict, so we preserve the ordering of the keys and
            # thus, formatting
            prl = json.load(f, object_pairs_hook=OrderedDict)
    except FileNotFoundError:
        # KiCAD didn't generate project local settings, let's create an empty one
        prl = {
            "board": {}
        }
    prl["board"]["visible_layers"] = "ffc000c_7ffffffe"
    prl["board"]["visible_items"] = [
        1,
        2,
        3,
        4,
        9,
        10,
        12,
        13,
        21,
        22,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        34,
        35
    ]
    with open(prlPath, "w", encoding="utf-8") as f:
        json.dump(prl, f, indent=2)
    pass

from pathlib import Path
import os

def create(inputboard, outputdir, jigsize, jigthickness, pcbthickness,
           registerborder, tolerance, ignore, cutout):
    board = pcbnew.LoadBoard(inputboard)
    refs = parseReferences(ignore)
    removeComponents(board, refs)

    Path(outputdir).mkdir(parents=True, exist_ok=True)

    jigsize = (fromMm(jigsize[0]), fromMm(jigsize[1]))
    addJigFrame(board, jigsize)
    cutoutComponents(board, getComponents(board, parseReferences(cutout)))

    stencilFile = os.path.join(outputdir, "stencil.kicad_pcb")
    board.Save(stencilFile)

    setStencilLayerVisibility(stencilFile)

    plotPlan = [
        # name, id, comment
        ("PasteBottom", pcbnew.B_Paste, "Paste Bottom"),
        ("PasteTop", pcbnew.F_Paste, "Paste top"),
    ]
    # get a copy of exportSettingsJlcpcb dictionary and
    # exclude the Edge.Cuts layer for creation of stencil gerber files
    exportSettings = exportSettingsJlcpcb.copy()
    exportSettings["ExcludeEdgeLayer"] = True
    gerberDir = os.path.join(outputdir, "gerber")
    gerberImpl(stencilFile, gerberDir, plotPlan, False, exportSettings)

    shutil.make_archive(os.path.join(outputdir, "gerbers"), "zip", gerberDir)

    jigthickness = fromMm(jigthickness)
    pcbthickness = fromMm(pcbthickness)
    outerBorder, innerBorder = fromMm(registerborder[0]), fromMm(registerborder[1])
    tolerance = fromMm(tolerance)
    topRegister = makeTopRegister(board, jigsize,jigthickness, pcbthickness,
        outerBorder, innerBorder, tolerance)
    bottomRegister = makeBottomRegister(board, jigsize,jigthickness, pcbthickness,
        outerBorder, innerBorder, tolerance)

    topRegisterFile = os.path.join(outputdir, "topRegister.scad")
    solid.scad_render_to_file(topRegister, topRegisterFile)
    renderScad(topRegisterFile, os.path.join(outputdir, "topRegister.stl"))

    bottomRegisterFile = os.path.join(outputdir, "bottomRegister.scad")
    solid.scad_render_to_file(bottomRegister, bottomRegisterFile)
    renderScad(bottomRegisterFile, os.path.join(outputdir, "bottomRegister.stl"))

def printedStencilSubstrate(outlineDxf, thickness, frameHeight, frameWidth, frameClearance):
    bodyOffset = solid.utils.up(0) if frameWidth + frameClearance == 0 else solid.offset(r=frameWidth + frameClearance)
    body = solid.linear_extrude(height=thickness + frameHeight)(
        bodyOffset(solid.import_dxf(outlineDxf)))
    boardOffset = solid.utils.up(0) if frameClearance == 0 else solid.offset(r=frameClearance)
    board = solid.utils.up(thickness)(
        solid.linear_extrude(height=thickness + frameHeight)(
            boardOffset(solid.import_dxf(outlineDxf))))
    return body - board

def getComponents(board, references):
    """
    Return a list of components based on designator
    """
    return [f for f in board.GetFootprints() if f.GetReference() in references]

def collectFootprintEdges(footprint, layer):
    """
    Return all edges on given layer in given footprint
    """
    return [e for e in footprint.GraphicalItems() if e.GetLayer() == layer]

def extractComponentPolygons(footprints, srcLayer):
    """
    Return a list of shapely polygons with holes for already placed components.
    The source layer defines the geometry on which the cutout is computed.
    Usually it a font or back courtyard
    """
    polygons = []
    for f in footprints:
       edges = collectFootprintEdges(f, srcLayer)
       for ring in extractRings(edges):
           polygons.append(toShapely(ring, edges))
    return polygons

def printedStencil(outlineDxf, holesDxf, extraHoles, thickness, frameHeight, frameWidth,
                   frameClearance, enlargeHoles, front):
    zScale = -1 if front else 1
    xRotate = 180 if front else 0
    substrate = solid.scale([1, 1, zScale])(printedStencilSubstrate(outlineDxf,
        thickness, frameHeight, frameWidth, frameClearance))
    holesOffset = solid.utils.up(0) if enlargeHoles == 0 else solid.offset(delta=enlargeHoles)
    holes = solid.linear_extrude(height=4*thickness, center=True)(
        holesOffset(solid.import_dxf(holesDxf)))
    substrate -= holes
    for h in extraHoles:
        substrate -= solid.scale([toMm(1), -toMm(1), 1])(
                solid.linear_extrude(height=4*thickness, center=True)(
                    solid.polygon(h.exterior.coords)))
    return solid.rotate(a=xRotate, v=[1, 0, 0])(substrate)

def createPrinted(inputboard, outputdir, pcbthickness, thickness, framewidth,
                  ignore, cutout, frameclearance, enlargeholes):
    """
    Create a 3D printed self-registering stencil.
    """
    board = pcbnew.LoadBoard(inputboard)
    refs = parseReferences(ignore)
    cutoutComponents = getComponents(board, parseReferences(cutout))
    removeComponents(board, refs)
    Path(outputdir).mkdir(parents=True, exist_ok=True)

    # We create the stencil based on DXF export. Using it avoids the necessity
    # to interpret KiCAD PAD shapes which constantly change with newer and newer
    # versions.
    height = min(pcbthickness, max(0.5, pcbthickness - 0.3))
    bottomPaste, topPaste, outline = pasteDxfExport(board, outputdir)
    # On Windows, OpenSCAD requires to use forward slashes instead of backslashes,
    # hence, the replacement:
    if os.name == "nt":
        bottomPaste = bottomPaste.replace("\\", "/")
        topPaste = topPaste.replace("\\", "/")
        outline = outline.replace("\\", "/")

    topCutout = extractComponentPolygons(cutoutComponents, pcbnew.F_CrtYd)
    bottomCutout = extractComponentPolygons(cutoutComponents, pcbnew.B_CrtYd)
    topStencil = printedStencil(outline, topPaste, topCutout, thickness, height,
        framewidth, frameclearance, enlargeholes, True)
    bottomStencil = printedStencil(outline, bottomPaste, bottomCutout, thickness,
        height, framewidth, frameclearance, enlargeholes, False)

    bottomStencilFile = os.path.join(outputdir, "bottomStencil.scad")
    solid.scad_render_to_file(bottomStencil, bottomStencilFile,
        file_header=f'$fa = 0.4; $fs = 0.4;', include_orig_code=True)
    renderScad(bottomStencilFile, os.path.join(outputdir, "bottomStencil.stl"))

    topStencilFile = os.path.join(outputdir, "topStencil.scad")
    solid.scad_render_to_file(topStencil, topStencilFile,
        file_header=f'$fa = 0.4; $fs = 0.4;', include_orig_code=True)
    renderScad(topStencilFile, os.path.join(outputdir, "topStencil.stl"))






import traceback
import sys
import click

from .common import execute_with_debug

def fabCommand(f):
    """
    A decorator to add the same functionality to all fab commands
    """
    # Note that the decorators has to be specified in a reverse order
    f = click.argument("outputdir", type=click.Path(file_okay=False))(f)
    f = click.argument("board", type=click.Path(dir_okay=False))(f)

    f = click.option('--drc/--no-drc', is_flag=True, default=True,
        help="Run DRC check before building the output.")(f)
    f = click.option("--nametemplate", default="{}",
        help="Template for naming the output files.")(f)
    f = click.option("--debug", is_flag=True, default=False,
        help="Print extra debugging information")(f)
    return f

@click.command()
@fabCommand
@click.option("--assembly/--no-assembly", help="Generate files for SMT assembly (schematics is required)")
@click.option("--schematic", type=click.Path(dir_okay=False), help="Board schematics (required for assembly files)")
@click.option("--ignore", type=str, default="", help="Comma separated list of designators to exclude from SMT assembly")
@click.option("--field", type=str, default="LCSC",
    help="Comma separated list of component fields field with LCSC order code. First existing field is used")
@click.option("--corrections", type=str, default="JLCPCB_CORRECTION",
    help="Comma separated list of component fields with the correction value. First existing field is used")
@click.option("--correctionpatterns", type=click.Path(dir_okay=False))
@click.option("--missingError/--missingWarn", help="If a non-ignored component misses LCSC field, fail")
@click.option("--autoname/--no-autoname", is_flag=True, help="Automatically name the output files based on the board name")
def jlcpcb(**kwargs):
    """
    Prepare fabrication files for JLCPCB including their assembly service
    """
    from kikit.fab import jlcpcb
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()
    return execute_with_debug(jlcpcb.exportJlcpcb, kwargs)

@click.command()
@fabCommand
@click.option("--assembly/--no-assembly", help="Generate files for SMT assembly (schematics is required)")
@click.option("--schematic", type=click.Path(dir_okay=False), help="Board schematics (required for assembly files)")
@click.option("--ignore", type=str, default="", help="Comma separated list of designators to exclude from SMT assembly")
@click.option("--corrections", type=str, default="PCBWAY_CORRECTION",
    help="Comma separated list of component fields with the correction value. First existing field is used")
@click.option("--correctionpatterns", type=click.Path(dir_okay=False))
@click.option("--manufacturer", type=str, default="Manufacturer",
    help="Comma separated list of fields to extract manufacturer name from. First existing field is used.")
@click.option("--partNumber", type=str, default="PartNumber",
    help="Comma separated list of fields to extract part number from. First existing field is used.")
@click.option("--description", type=str, default="Description",
    help="Comma separated list of fields to extract description from. First existing field is used.")
@click.option("--notes", type=str, default="Notes",
    help="Comma separated list of fields to extract notes from. First existing field is used.")
@click.option("--solderType", type=str, default="Type",
    help="Comma separated list of fields to extract solder type from. First existing field is used.")
@click.option("--footprint", type=str, default="FootprintPCBWay",
    help="Comma separated list of fields to extract the footprint name for the BOM. First existing field is used, otherwise the footprint library name.")
@click.option("--nBoards", type=int, default=1,
    help="Number of boards per panel (default 1).")
@click.option("--missingError/--missingWarn", help="If a non-ignored component misses Manufacturer / PartNumber field, fail")
def pcbway(**kwargs):
    """
    Prepare fabrication files for PCBWAY including their assembly service
    """
    from kikit.fab import pcbway
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()
    return execute_with_debug(pcbway.exportPcbway, kwargs)


@click.command()
@fabCommand
def oshpark(**kwargs):
    """
    Prepare fabrication files for OSH Park
    """
    from kikit.fab import oshpark
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()
    return execute_with_debug(oshpark.exportOSHPark, kwargs)

@click.command()
@fabCommand
@click.option("--schematic", type=click.Path(dir_okay=False), help="Board schematics (required for assembly files)")
@click.option("--ignore", type=str, default="", help="Comma separated list of designators to exclude from SMT assembly")
@click.option("--corrections", type=str, default="YY1_CORRECTION",
    help="Comma separated list of component fields with the correction value. First existing field is used")
@click.option("--correctionpatterns", type=click.Path(dir_okay=False))
def neodenyy1(**kwargs):
    """
    Prepare fabrication files for Neoden YY1
    """
    from kikit.fab import neodenyy1
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()
    return execute_with_debug(neodenyy1.exportNeodenYY1, kwargs)

@click.command()
@fabCommand
def openpnp(**kwargs):
    """
    Prepare fabrication files for OpenPnP
    """
    from kikit.fab import openpnp
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()
    return execute_with_debug(openpnp.exportOpenPnp, kwargs)

@click.group()
def fab():
    """
    Export complete manufacturing data for given fabrication houses
    """
    pass

fab.add_command(jlcpcb)
fab.add_command(pcbway)
fab.add_command(oshpark)
fab.add_command(neodenyy1)
fab.add_command(openpnp)



import datetime as dt
from string import Template
from typing import Callable, Optional, Dict, Any
import pcbnew


class Formatter:
    """
    Turns a function into a formatter. Caches the result.
    """
    def __init__(self, fn: Callable[[], str], vars: Dict[str, str]={}) -> None:
        self.fn = fn
        self.value: Optional[str] = None
        self.vars = vars

    def __str__(self) -> str:
        if self.value is None:
            self.value = self.expandVariables(self.fn())
        return self.value

    def expandVariables(self, string: str) -> str:
        try:
            return Template(string).substitute(self.vars)
        except KeyError as e:
            raise RuntimeError(f"Requested text '{string}' expects project variable '{e}' which is missing") from None

def kikitTextVars(board: pcbnew.BOARD, vars: Dict[str, str]={}) -> Dict[str, Any]:
    availableVars: Dict[str, Formatter] = {
        "date": Formatter(lambda: dt.datetime.today().strftime("%Y-%m-%d"), vars),
        "time24": Formatter(lambda: dt.datetime.today().strftime("%H:%M"), vars),
        "year": Formatter(lambda: dt.datetime.today().strftime("%Y"), vars),
        "month": Formatter(lambda: dt.datetime.today().strftime("%m"), vars),
        "day": Formatter(lambda: dt.datetime.today().strftime("%d"), vars),
        "hour": Formatter(lambda: dt.datetime.today().strftime("%H"), vars),
        "minute": Formatter(lambda: dt.datetime.today().strftime("%M"), vars),
        "second": Formatter(lambda: dt.datetime.today().strftime("%S"), vars),
        "boardTitle": Formatter(lambda: board.GetTitleBlock().GetTitle(), vars),
        "boardDate": Formatter(lambda: board.GetTitleBlock().GetDate(), vars),
        "boardRevision": Formatter(lambda: board.GetTitleBlock().GetRevision(), vars),
        "boardCompany": Formatter(lambda: board.GetTitleBlock().GetCompany(), vars)
    }

    for i in range(10):
        availableVars[f"boardComment{i + 1}"] = Formatter(lambda x = i: board.GetTitleBlock().GetComment(x), vars)

    for var, value in vars.items():
        availableVars[f"user-{var}"] = value

    return availableVars



import click

@click.command()
@click.argument("outdir", type=click.Path(file_okay=False))
@click.option("--description", "-d", type=click.Path(dir_okay=False),
    required=True, help="markdown file with page text")
@click.option("--board", "-b", type=(str, str, click.Path(dir_okay=False)),
    multiple=True, help="<name> <comment> <kicad_pcb file> to include in generated page.")
@click.option("--resource", "-r", type=click.Path(dir_okay=True), multiple=True,
    help="Additional resource files to (e.g., images referenced in description) to include.")
@click.option("--template", type=click.Path(), default="default",
    help="Path to a template directory or a name of built-in one. See doc/present.md for template specification.")
@click.option("--repository", type=str, help="URL of the repository")
@click.option("--name", type=str, help="Name of the board (used e.g., for title)", required=True)
def boardpage(**kwargs):
    """
    Build a board presentation page based on markdown description and include
    download links for board sources and gerbers.
    """
    from kikit import present
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()

    return present.boardpage(**kwargs)

@click.group()
def present():
    """
    Prepare board presentation
    """
    pass

present.add_command(boardpage)



import wx
import threading

app = None

def initDialog(initFn):
    global app
    try:
        return initFn()
    except Exception:
        try:
            # Some Linux distributions ship incompatible wxPhoenix version.
            # So we start a custom wxApp loop
            app = wx.App()
            app.InitLocale()
            t = threading.Thread(target=app.MainLoop)
            t.daemon = True
            t.start()
            return initFn()
        except Exception as e:
            raise e from None

def destroyDialog(dialog):
    if dialog is not None:
        dialog.Destroy()



import click
from kikit import (panelize_ui, export_ui, present_ui, stencil_ui,
    modify_ui, fab_ui, drc_ui)
from kikit import __version__
import sys

@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(__version__)
def cli():
    pass

cli.add_command(export_ui.export)
cli.add_command(panelize_ui.panelize)
cli.add_command(panelize_ui.separate)
cli.add_command(present_ui.present)
cli.add_command(modify_ui.modify)
cli.add_command(stencil_ui.stencil)
cli.add_command(fab_ui.fab)
cli.add_command(drc_ui.drc)


if __name__ == '__main__':
    # When KiCAD crashes, we want the user to know
    import faulthandler
    import sys
    faulthandler.enable(sys.stderr)

    cli()



import click
import sys

from .common import execute_with_debug


@click.command()
@click.argument("inputBoard", type=click.Path(dir_okay=False))
@click.argument("outputDir", type=click.Path(dir_okay=True))
@click.option("--pcbthickness", type=float, default=1.6,
    help="PCB thickness in mm")
@click.option("--thickness", type=float, default=0.15,
    help="Stencil thickness in mm. Defines amount of paste dispensed")
@click.option("--framewidth", type=float, default=1,
    help="Register frame width")
@click.option("--ignore", type=str, default="",
    help="Comma separated list of components references to exclude from the stencil")
@click.option("--cutout", type=str, default="",
    help="Comma separated list of components references to cutout from the stencil based on the courtyard")
@click.option("--frameclearance", type=float, default=0,
    help="Clearance for the stencil register in milimeters")
@click.option("--enlargeholes", type=float, default=0,
    help="Enlarge pad holes by x mm")
@click.option("--debug", is_flag=True, default=False,
        help="Print extra debugging information")
def createPrinted(**kwargs):
    """
    Create a 3D printed self-registering stencil.
    """
    from kikit import stencil

    return execute_with_debug(stencil.createPrinted, kwargs)


@click.command()
@click.argument("inputBoard", type=click.Path(dir_okay=False))
@click.argument("outputDir", type=click.Path(dir_okay=True))
@click.option("--jigsize", type=(int, int), default=(100, 100),
    help="Jig frame size in mm: <width> <height>")
@click.option("--jigthickness", type=float, default=3,
    help="Jig thickness in mm")
@click.option("--pcbthickness", type=float, default=1.6,
    help="PCB thickness in mm")
@click.option("--registerborder", type=(float, float), default=(3, 1),
    help="Register borders in mm: <outer> <inner>")
@click.option("--tolerance", type=float, default=0.05,
    help="Enlarges the register by the tolerance value")
@click.option("--ignore", type=str, default="",
    help="Comma separated list of components references to exclude from the stencil")
@click.option("--cutout", type=str, default="",
    help="Comma separated list of components references to cutout from the stencil based on the courtyard")
@click.option("--debug", is_flag=True, default=False,
        help="Print extra debugging information")
def create(**kwargs):
    """
    Create stencil and register elements for manual paste dispensing jig.
    See more details at: https://github.com/yaqwsx/KiKit/blob/master/doc/stencil.md
    """
    from kikit import stencil
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()

    return execute_with_debug(stencil.create, kwargs)


@click.group()
def stencil():
    """
    Create solder paste stencils
    """
    pass

stencil.add_command(create)
stencil.add_command(createPrinted)



import inspect
import tempfile
import subprocess
import shutil
import os
import sys
import threading
from itertools import chain

def header(func):
    signature = inspect.signature(func)
    args = [str(k) if v.default is inspect.Parameter.empty else str(k) + "=" + str(v.default)
        for k, v in signature.parameters.items()]
    currentLine = func.__name__ + "("
    lines = []
    indent = " " * (len(func.__name__) + 1)
    separator = ""
    for arg in args:
        newLine = currentLine + separator + arg
        if len(newLine) > 80:
            lines.append(currentLine + separator)
            currentLine = indent + arg
        else:
            currentLine = newLine
        separator = ", "
    lines.append(currentLine + ")")
    return("\n".join(lines))

def printHeader(func):
    print(f"\n#### `{func.__name__}`")
    print(f"```\n{header(func)}\n```")

def printHelp(x):
    print(inspect.getdoc(x))

def quotePosix(args):
    """
    Given a list of command line arguments, quote them so they can be can be
    printed on POSIX
    """
    def q(x):
        if " " in x:
            return "'" + x + "'"
        else:
            return x
    return [q(x) for x in args]

def quoteWindows(args):
    """
    Given a list of command line arguments, quote them so they can be can be
    printed in Windows CLI
    """
    def q(x):
        if " " in x:
            return '"' + x + '"'
        else:
            return x
    return [q(x) for x in args]

def panelizeAndDraw(name, command):
    dirname = tempfile.mkdtemp()
    output = os.path.join(dirname, "x.kicad_pcb")
    try:
        outimage = f"docs/resources/{name}.png"
        subprocess.run(command + [output, "--debug", "trace: true"], check=True, capture_output=True)
        r = subprocess.run(["pcbdraw", "plot", "--vcuts", "Cmts.User", "--silent", output,
                outimage], check=True, capture_output=True)
        subprocess.run(["convert", outimage, "-define",
            "png:include-chunk=none", outimage], check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        print("Command: " + " ".join(e.cmd), file=sys.stderr)
        print("Stdout: " + e.stdout.decode("utf8"), file=sys.stderr)
        print("Stderr: " + e.stderr.decode("utf8"), file=sys.stderr)
        sys.exit(1)
    finally:
        shutil.rmtree(dirname)


runExampleThreads = []

def runBoardExample(name, args):
    """
    Run kikit CLI command with args - omitting the output file name. Prints a
    markdown with the command and includes a generated board image.

    Generating images runs in parallel, so do not forget to invoke
    runBoardExampleJoin() at the end of your script.
    """
    realArgs = ["python3", "-m", "kikit.ui"] + list(chain(*args))

    # We print first, so in a case of failure we have the command in a nice
    # copy-paste-ready form
    args[0] = ["kikit"] + args[0]
    args[-1] = args[-1] + ["panel.kicad_pcb"]

    print('!!! example "Panelization command"\n')

    print('    === "Linux/macOS"')
    print("        ```")
    for i, c in enumerate(args):
        if i == 0:
            print("        ", end="")
        else:
            print("            ", end="")
        end = "\n" if i + 1 == len(args) else " \\\n"
        print(" ".join(quotePosix(c)), end=end)
    print("        ```")
    print("")
    print('    === "Windows"')
    print("        ```")
    for i, c in enumerate(args):
        if i == 0:
            print("        ", end="")
        else:
            print("            ", end="")
        end = "\n" if i + 1 == len(args) else " ^\n"
        print(" ".join(quoteWindows(c)), end=end)
    print("        ```\n")
    print("    ![{0}](../resources/{0}.png)".format(name))

    t = threading.Thread(target=lambda: panelizeAndDraw(name, realArgs))
    t.start()
    global runExampleThreads
    runExampleThreads.append(t)

def runScriptingExample(name, args):
    """
    Run a Python panelization script that takes the name of the output as a last
    argument and create a drawing of it.
    """

    realArgs = ["python3"] + list(chain(*args))
    print("```")
    for i, c in enumerate(args):
        if i != 0:
            print("    ", end="")
        end = "\n" if i + 1 == len(args) else " \\\n"
        print(" ".join(quote(c)), end=end)
    print("```\n")
    print("![{0}](../resources/{0}.png)".format(name))

    t = threading.Thread(target=lambda: panelizeAndDraw(name, realArgs))
    t.start()
    global runExampleThreads
    runExampleThreads.append(t)


def runExampleJoin():
    for t in runExampleThreads:
        t.join()




from typing import Set
import pcbnew
from .defs import Layer
import re

def references(board: pcbnew.BOARD, show: bool, pattern: str,
               allowedLayers: Set[Layer] = set(Layer.all()),
               isSelectedItemOnly: bool = False) -> None:
    """
    Show or hide references in a footprint.
    """
    for footprint in board.GetFootprints():
        if isSelectedItemOnly and not footprint.IsSelected():
            continue
        if re.match(pattern, footprint.GetReference()) and footprint.Reference().GetLayer() in allowedLayers:
            footprint.Reference().SetVisible(show)
        for x in footprint.GraphicalItems():
            if not isinstance(x, pcbnew.PCB_FIELD) and not isinstance(x, pcbnew.PCB_TEXT):
                continue
            if x.GetText().strip() in ["${REFERENCE}", "REF**"] and x.GetLayer() in allowedLayers:
                x.SetVisible(show)


def values(board: pcbnew.BOARD, show: bool, pattern: str,
           allowedLayers: Set[Layer] = set(Layer.all()),
           isSelectedItemOnly: bool = False) -> None:
    """
    Show or hide values in a footprint.
    """
    for footprint in board.GetFootprints():
        if isSelectedItemOnly and not footprint.IsSelected():
            continue
        if re.match(pattern, footprint.GetReference()) and footprint.Value().GetLayer() in allowedLayers:
            footprint.Value().SetVisible(show)
        for x in footprint.GraphicalItems():
            if not isinstance(x, pcbnew.PCB_FIELD) and not isinstance(x, pcbnew.PCB_TEXT):
                continue
            if x.GetText().strip() in ["${VALUE}", "VAL**"] and x.GetLayer() in allowedLayers:
                x.SetVisible(show)



from __future__ import annotations
import sys
import traceback
from typing import List, Optional, Tuple, Union, Callable
from kikit.defs import Layer
from kikit.typing import Box
import pcbnew
from pcbnew import BOX2I, VECTOR2I, EDA_ANGLE
from kikit.intervals import AxialLine
import os
from itertools import product, chain, islice
import numpy as np
from shapely.geometry import LinearRing
import shapely.geometry

PKG_BASE = os.path.dirname(__file__)
KIKIT_LIB = os.path.join(PKG_BASE, "resources/kikit.pretty")
SHP_EPSILON = pcbnew.FromMM(0.001) # Common factor of enlarging substrates to
                                   # cover up numerical imprecisions of Shapely

KiLength = int
KiAngle = EDA_ANGLE
KiPoint = VECTOR2I

def fromDegrees(angle: Union[float,int]) -> KiAngle:
    """Convert angle in degrees to Kicad angle representation"""
    return EDA_ANGLE(angle, pcbnew.DEGREES_T)

def fromMm(mm: float) -> KiLength:
    """Convert millimeters to KiCAD internal units"""
    return pcbnew.FromMM(mm)

def toMm(kiUnits: KiLength) -> float:
    """Convert KiCAD internal units to millimeters"""
    return pcbnew.ToMM(int(kiUnits))

def toKiCADPoint(p) -> KiPoint:
    """Convert tuple or array like objects to KiCAD point (VECTOR2I)"""
    assert len(p) == 2
    return VECTOR2I(*[int(x) for x in p])

def fitsIn(what: Union[BOX2I, VECTOR2I], where: BOX2I) -> bool:
    """
    Return true iff 'what' (BOX2I or VECTOR2I) is fully contained in 'where'
    (BOX2I)
    """
    assert isinstance(what, (BOX2I, VECTOR2I, pcbnew.wxPoint))
    if isinstance(what, (VECTOR2I, pcbnew.wxPoint)):
        return (what[0] >= where.GetX() and
                what[0] <= where.GetX() + where.GetWidth() and
                what[1] >= where.GetY() and
                what[1] <= where.GetY() + where.GetHeight())
    else:
        return (what.GetX() >= where.GetX() and
                what.GetX() + what.GetWidth() <= where.GetX() + where.GetWidth() and
                what.GetY() >= where.GetY() and
                what.GetY() + what.GetHeight() <= where.GetY() + where.GetHeight())

def combineBoundingBoxes(a, b):
    """ Retrun BOX2I as a combination of source bounding boxes """
    x1 = min(a.GetX(), b.GetX())
    y1 = min(a.GetY(), b.GetY())
    x2 = max(a.GetX() + a.GetWidth(), b.GetX() + b.GetWidth())
    y2 = max(a.GetY() + a.GetHeight(), b.GetY() + b.GetHeight())
    return BOX2I(toKiCADPoint((x1, y1)), toKiCADPoint((x2 - x1, y2 - y1)))

def collectEdges(board, layerId, sourceArea=None):
    """ Collect edges in sourceArea on given layer including footprints """
    edges = []
    for edge in chain(board.GetDrawings(), *[m.GraphicalItems() for m in board.GetFootprints()]):
        if edge.GetLayer() != layerId:
            continue
        if isinstance(edge, pcbnew.PCB_DIMENSION_BASE):
            continue
        if isinstance(edge, pcbnew.PCB_TEXT):
            continue
        if not sourceArea or fitsIn(edge.GetBoundingBox(), sourceArea):
            edges.append(edge)
    return edges

def collectItems(boardCollection, sourceArea):
    """ Returns a list of board items fully contained in the source area """
    return list([x for x in boardCollection if fitsIn(x.GetBoundingBox(), sourceArea)])

def collectFootprints(boardCollection, sourceArea):
    """
    Returns a list of board footprints Which origin fits inside the source area.
    """
    return list([x for x in boardCollection if fitsIn(x.GetPosition(), sourceArea)])

def collectZones(boardCollection, sourceArea):
    """
    Returns a list of board zones which centroid fits inside the source area.
    """
    def zoneCentroid(zone: pcbnew.ZONE) -> pcbnew.VECTOR2I:
        items = []
        for outline in [zone.Outline().Outline(i) for i in range(zone.Outline().OutlineCount())]:
            p = shapely.geometry.Polygon([outline.CPoint(i) for i in range(outline.PointCount())])
            items.append(p)
        polygon = shapely.ops.unary_union(items)
        return pcbnew.VECTOR2I(*[int(x) for x in polygon.centroid.coords[0]])
    return list([x for x in boardCollection if fitsIn(zoneCentroid(x), sourceArea)])

def getBBoxWithoutContours(edge):
    width = edge.GetWidth()
    edge.SetWidth(0)
    bBox = edge.GetBoundingBox()
    edge.SetWidth(width)
    return bBox

def listGeometries(shapelyObject):
    """
    Given a shapely object, return an iterable of all geometries. I.e., for
    single items, return an iterable containing only the original item. For
    collections, return iterable of all the geometries in it.
    """
    if hasattr(shapelyObject, 'geoms'):
        return shapelyObject.geoms
    return [shapelyObject]

def findBoundingBox(edges):
    """
    Return a bounding box of all drawings in edges
    """
    if len(edges) == 0:
        raise RuntimeError("No board edges found")
    boundingBox = getBBoxWithoutContours(edges[0])
    for edge in edges[1:]:
        boundingBox = combineBoundingBoxes(boundingBox, getBBoxWithoutContours(edge))
    return boundingBox

def findBoardBoundingBox(board, sourceArea=None):
    """
    Returns a bounding box (BOX2I) of all Edge.Cuts items either in
    specified source area (BOX2I) or in the whole board
    """
    edges = collectEdges(board, Layer.Edge_Cuts, sourceArea)
    return findBoundingBox(edges)

def rectCenter(rect):
    """
    Given a BOX2I return its center
    """
    return toKiCADPoint((rect.GetX() + rect.GetWidth() // 2, rect.GetY() + rect.GetHeight() // 2))

def rectByCenter(center, width, height):
    """
    Given a center point and size, return BOX2I
    """
    return BOX2I(
        toKiCADPoint((center[0] - width // 2, center[1] - height // 2)),
        toKiCADPoint((width, height)))

def normalize(vector):
    """ Return a vector with unit length """
    vec = np.array([vector[0], vector[1]])
    return vec / np.linalg.norm(vector)

def makePerpendicular(vector):
    """
    Given a 2D vector, return a vector which is perpendicular to the input one
    """
    return np.array([vector[1], -vector[0]])

def linestringToSegments(linestring):
    """
    Given a Shapely linestring, return a list of tuples with start and endpoint
    of the segment
    """
    return [x for x in zip(linestring.coords, islice(linestring.coords, 1, None))]

def tl(rect):
    """ Return top left corner of rect """
    return toKiCADPoint((rect.GetX(), rect.GetY()))

def tr(rect):
    """ Return top right corner of rect """
    return toKiCADPoint((rect.GetX() + rect.GetWidth(), rect.GetY()))

def br(rect):
    """ Return bottom right corner of rect """
    return toKiCADPoint((rect.GetX() + rect.GetWidth(), rect.GetY() + rect.GetHeight()))

def bl(rect):
    """ Return bottom left corner of rect """
    return toKiCADPoint((rect.GetX(), rect.GetY() + rect.GetHeight()))

def removeComponents(board, references):
    """
    Remove components with references from the board. References is a list of
    strings
    """
    for footprint in board.GetFootprints():
        if footprint.GetReference() in references:
            board.Remove(footprint)

def parseReferences(dStr):
    """
    Parse comma separated list of component references to a list
    """
    return [x.strip() for x in dStr.split(",") if len(x.strip()) > 0]


def shpBBoxLeft(bbox):
    """
    Given a shapely bounding box, return left edge as (pos, interval)
    """
    return AxialLine(bbox[0], bbox[1], bbox[3])

def shpBBoxRight(bbox):
    """
    Given a shapely bounding box, return right edge as (pos, interval)
    """
    return AxialLine(bbox[2], bbox[1], bbox[3])

def shpBBoxTop(bbox):
    """
    Given a shapely bounding box, return top edge as (pos, interval)
    """
    return AxialLine(bbox[1], bbox[0], bbox[2])

def shpBBoxBottom(bbox):
    """
    Given a shapely bounding box, return bottom edge as (pos, interval)
    """
    return AxialLine(bbox[3], bbox[0], bbox[2])

def shpBBoxMerge(a: Box, b: Box) -> Box:
    """
    Given two shapely bounding boxes, return smallest bounding box where both
    can fit.
    """
    return (
        min(a[0], b[0]),
        min(a[1], b[1]),
        max(a[2], b[2]),
        max(a[3], b[3])
    )

def shpBBoxExpand(box: Box, x: float, y: Optional[float]=None) -> Box:
    """
    Given a shapely bounding box, return new one expanded by given amount. If y
    is not supplied, it the same as x.
    """
    if y is None:
        y = x
    return (box[0] - x, box[1] - y, box[2] + x, box[3] + y)

def shpBoxToRect(box):
    box = list([int(x) for x in box])
    return BOX2I(toKiCADPoint((box[0], box[1])),
                 toKiCADPoint((box[2] - box[0], box[3] - box[1])))

def rectToShpBox(rect):
    return shapely.geometry.box(rect.GetX(), rect.GetY(),
        rect.GetX() + rect.GetWidth(), rect.GetY() + rect.GetHeight())

def isLinestringCyclic(line):
    c = line.coords
    return c[0] == c[-1] or isinstance(line, LinearRing)

def constructArrow(origin, direction, distance: float, tipSize: float) -> shapely.LineString:
    origin = np.array(origin)
    direction = np.array(direction)

    endpoint = origin + direction * distance

    tipEndpoint1 = endpoint + tipSize / 2 * (-direction - np.array([-direction[1], direction[0]]))
    tipEndpoint2 = endpoint + tipSize / 2 * (-direction + np.array([-direction[1], direction[0]]))

    arrow = shapely.LineString([origin, endpoint, tipEndpoint1, tipEndpoint2, endpoint])
    return arrow

def fromOpt(object, default):
    """
    Given an object, return it if not None. Otherwise return default
    """
    return object if object is not None else default

def isBottomLayer(layer):
    """
    Decide if layer is a bottom layer
    """
    return layer.name.startswith("B_")

def commonPoints(lines):
    """
    Given a list of lines, return dictionary - vertice -> count. Where count
    specifies how many lines share the vertex.
    """
    count = {}
    for l in lines:
        for c in l.coords:
            count[c] = count.get(c, 0) + 1
    return count

def isHorizontal(start, end):
    """
    Given a line decide if it is horizontal
    """
    return start[1] == end[1]

def isVertical(start, end):
    """
    Given a line decide if it is vertical
    """
    return start[0] == end[0]

def resolveAnchor(anchor):
    """
    Given a string anchor name, return a function that transforms BOX2I into
    a VECTOR2I
    """
    choices = {
        "tl": lambda x: x.GetPosition() + toKiCADPoint((0, 0)),
        "tr": lambda x: x.GetPosition() + toKiCADPoint((x.GetWidth(), 0)),
        "bl": lambda x: x.GetPosition() + toKiCADPoint((0, x.GetHeight())),
        "br": lambda x: x.GetPosition() + toKiCADPoint((x.GetWidth(), x.GetHeight())),
        "mt": lambda x: x.GetPosition() + toKiCADPoint((x.GetWidth() / 2, 0)),
        "mb": lambda x: x.GetPosition() + toKiCADPoint((x.GetWidth() / 2, x.GetHeight())),
        "ml": lambda x: x.GetPosition() + toKiCADPoint((0, x.GetHeight() / 2)),
        "mr": lambda x: x.GetPosition() + toKiCADPoint((x.GetWidth(), x.GetHeight() / 2)),
        "c":  lambda x: x.GetPosition() + toKiCADPoint((x.GetWidth() / 2, x.GetHeight() / 2))
    }
    return choices[anchor]

def splitOn(input: str, predicate: Callable[[str], bool]) \
        -> Tuple[str, str]:
    """
    Split a string into a head fullfilling predicate and the rest
    """
    left = ""
    for i, x in enumerate(input):
        if predicate(x):
            left += x
        else:
            break
    return left, input[i:]

def splitOnReverse(input: str, predicate: Callable[[str], bool]) \
        -> Tuple[str, str]:
    """
    Split a string into a tail fullfilling predicate and the remaning head
    """
    tail, head = splitOn(input[::-1], predicate)
    return head[::-1], tail[::-1]

def indexOf(list, predicate):
    """
    Return the index of the first element that satisfies predicate. If no
    element is found, return -1
    """
    for i, x in enumerate(list):
        if predicate(x):
            return i
    return -1

def readParameterList(inputStr):
    """
    Given a string, read semicolon separated parameter list in the form of
    `key: value; key: value`. You can escape via `\\`
    """
    from kikit.panelize_ui import splitStr

    if len(inputStr.strip()) == 0:
        return {}
    try:
        values = {}
        for i, pair in enumerate(splitStr(";", "\\", inputStr)):
            if len(pair.strip()) == 0:
                continue
            s = pair.split(":")
            if i == 0 and len(s) == 1:
                values["type"] = s[0].strip()
                continue
            key, value = s[0].strip(), s[1].strip()
            values[key] = value
        return values
    except (TypeError, IndexError):
        raise RuntimeError(f"'{pair}' is not a valid key: value pair")

def fakeKiCADGui():
    """
    KiCAD assumes wxApp and locale exists. If we invoke a command, fake the
    existence of an app. You should store the application in a top-level
    function of the command
    """
    import wx
    import os

    if hasattr(wx, "DisableAsserts"):
        wx.DisableAsserts()

    if os.name != "nt" and os.environ.get("DISPLAY", "").strip() == "":
        return None

    # Originally, we created wx App and initiated the locale. However, the only
    # purpose of this was to supress the warnings from KiCAD. It seems that the
    # existence of partially initiated app breaks some functions (e.g., the
    # Excellon writer). Therefore, instead of initiating the app, as shown
    # below, we simply redirect stdout/stderr to /dev/null and allow only for
    # output via sys.stdout and sys.stderr.
    # app = wx.App()
    # app.InitLocale()
    # return app

    sys.stdout = os.fdopen(os.dup(1), "w")
    sys.stderr = os.fdopen(os.dup(2), "w")

    os.dup2(os.open(os.devnull,os.O_RDWR), 1)
    os.dup2(os.open(os.devnull,os.O_RDWR), 2)

    return None

def execute_with_debug(procedure, kwargs):
    debug = kwargs["debug"]
    del kwargs["debug"]

    if debug:
        traceback.print_exc(file=sys.stderr)

    try:
        return procedure(**kwargs)
    except Exception as e:
        sys.stderr.write(f"An error occurred: {e}\n")
        sys.stderr.write("No output files produced\n")
        if debug:
            raise e from None
        sys.exit(1)



from typing import Union, Optional, Dict, Any
from pathlib import Path
from string import Template
from functools import cached_property
import pcbnew
import json


class KiCADProject:
    """
    This code represents a KiCAD project and allows easy access to individual
    files without the need to hassle with project names.
    """
    def __init__(self, path: Union[str, Path]) -> None:
        self.projectdir: Path = Path(path).resolve()
        name = None
        if str(path).endswith(".kicad_pro"):
            name = self.projectdir.name[:-len(".kicad_pro")]
            self.projectdir = self.projectdir.parent
        elif str(path).endswith(".kicad_pcb"):
            name = self.projectdir.name[:-len(".kicad_pcb")]
            self.projectdir = self.projectdir.parent
        else:
            if not self.projectdir.is_dir():
                raise RuntimeError(f"The project directory {self.projectdir} is not a directory")
            name = self._resolveProjectName(self.projectdir)
        self._name: str = name

    @staticmethod
    def _resolveProjectName(path: Path) -> str:
        candidate: Optional[str] = None
        for item in path.iterdir():
            if not item.name.endswith(".kicad_pro"):
                continue
            if candidate is not None:
                raise RuntimeError(f"There are multiple projects ({candidate} " +
                                   f"and {item.name}) in directory {path}. Not " +
                                   f"clear which one to choose.")
            candidate = item.name
        if candidate is not None:
            return candidate[:-len(".kicad_pro")]
        raise RuntimeError(f"No project found in {path}")

    @property
    def projectPath(self) -> Path:
        return self.projectdir / f"{self._name}.kicad_pro"

    @property
    def boardPath(self) -> Path:
        return self.projectdir / f"{self._name}.kicad_pcb"

    @property
    def schemaPath(self) -> Path:
        return self.projectdir / f"{self._name}.kicad_sch"

    @property
    def dirPath(self) -> Path:
        return self.projectdir

    def has(self, file: Union[str, Path]):
        return (self.projectdir / file).exists()

    def expandText(self, text: str):
        """
        Given KiCAD text expand
        """
        try:
            return Template(text).substitute(self.textVars)
        except KeyError as e:
            raise RuntimeError(f"Requested text '{text}' expects project variable '{e}' which is missing") from None

    @property
    def board(self) -> pcbnew.BOARD:
        return pcbnew.LoadBoard(str(self.getBoard()))

    @cached_property
    def projectJson(self) -> Dict[str, Any]:
        with open(self.projectPath, "r", encoding="utf-8") as f:
            return json.load(f)

    @cached_property
    def textVars(self) -> Dict[str, str]:
        return self.projectJson.get("text_variables", {})



from dataclasses import dataclass

@dataclass
class KiKitActionPlugin:
    package: str
    name: str
    description: str

availablePlugins = [
    KiKitActionPlugin("hideReferences", "Show/hide references",
        "Allows you to batch show or hide references based on regular expression"),
    KiKitActionPlugin("panelize", "Panelize design",
        "Allows you to specify panelization process via GUI")
]

def importAllPlugins():
    """
    Bring all plugins that KiKit offers into a the global namespace. This
    function is impure as it modifies the global variable scope. The purpose
    of this function is to allow the PCM proxy to operate.
    """
    import importlib

    for plugin in availablePlugins:
        module = importlib.import_module(f"kikit.actionPlugins.{plugin.package}")
        module.plugin().register()



from kikit import panelize
from kikit.panel_features.copperFill import HatchedCopperFill, HexCopperFill, SolidCopperFill
from kikit.panelize_ui import Section, PresetError
from kikit.panelize import *
from kikit.defs import Layer
from shapely.geometry import box
from kikit.plugin import HookPlugin
from kikit.text import kikitTextVars
from kikit.units import BaseValue, BaseAngle, PercentageValue
from kikit.panelize_ui_sections import *
from kikit.substrate import SubstrateNeighbors
from kikit.common import resolveAnchor
import commentjson
import enum
import json
import csv
import io

# This package exists as it is not needed for showing help and running other
# commands, however, it has heavy dependencies (pcbnew) that take a second
# to load. Splitting it speeds up KiKit's startup.

PKG_BASE = os.path.dirname(__file__)
PRESET_LIB = os.path.join(PKG_BASE, "resources/panelizePresets")


def encodePreset(value):
    """
    Convert a preset into its stringified version.
    """
    if hasattr(value, "__kikit_preset_repr"):
        return getattr(value, "__kikit_preset_repr")
    if value is None:
        return "none"
    if isinstance(value, BaseValue) or isinstance(value, BaseAngle):
        return str(value)
    if isinstance(value, PercentageValue):
        return str(value)
    if isinstance(value, EDA_TEXT_HJUSTIFY_T) or isinstance(value, EDA_TEXT_VJUSTIFY_T):
        return writeJustify(value)
    if isinstance(value, Enum):
        return str(value.name)
    if isinstance(value, bool):
        return value
    if isinstance(value, int):
        return value
    if isinstance(value, str):
        return value
    if isinstance(value, list):
        return ",".join([encodePreset(x) for x in value])
    if isinstance(value, dict):
        return {encodePreset(k): encodePreset(v) for k, v in value.items()}
    if isinstance(value, FootprintId):
        return f"{value.lib}:{value.footprint}"
    raise RuntimeError(f"Cannot serialize {value} of type {type(value)}")

def dumpPreset(preset):
    """
    Pretty prints a preset into string (containing JSON) so that it can be
    loaded back again by KiKit
    """
    # You might be wondering, why don't ve subclass JSONEncoder and instead, we
    # preprocess the preset. The reason for that is that json.dumps ignores
    # encoder for primitive types, which breaks BaseValue
    preprocessedPreset = encodePreset(preset)
    return json.dumps(preprocessedPreset, indent=4)

def validatePresetLayout(preset):
    if not isinstance(preset, dict):
        raise PresetError("Preset is not a dictionary")
    for name, section in preset.items():
        if not isinstance(section, dict):
            raise PresetError(f"Section '{name}' is not a dictionary")

def writeJustify(j):
    choices = {
        EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_LEFT: "left",
        EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_RIGHT: "right",
        EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_CENTER: "center",
        EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_TOP: "top",
        EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_CENTER: "center",
        EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_BOTTOM: "bottom"
    }
    return choices[j]

def postProcessPreset(preset):
    process = {
        "layout": ppLayout,
        "source": ppSource,
        "tabs": ppTabs,
        "cuts": ppCuts,
        "framing": ppFraming,
        "tooling": ppTooling,
        "fiducials": ppFiducials,
        "text": ppText,
        "text2": ppText,
        "text3": ppText,
        "text4": ppText,
        "copperfill": ppCopper,
        "post": ppPost,
        "page": ppPage,
        "debug": ppDebug
    }
    for name, section in preset.items():
        process[name](section)

def loadPreset(path):
    """
    Load a preset from path and perform simple validation on its structure.
    Automatically resolves built-in styles (prefixed with :, omitting suffix).
    """
    if path.startswith(":"):
        presetName = path
        path = os.path.join(PRESET_LIB, path[1:] + ".json")
        if not os.path.exists(path):
            raise RuntimeError(f"Uknown built-in preset '{presetName}'")
    try:
        with open(path, "r", encoding="utf-8") as f:
            preset = commentjson.load(f)
            validatePresetLayout(preset)
            return preset
    except OSError as e:
        raise RuntimeError(f"Cannot open preset '{path}'")
    except PresetError as e:
        raise PresetError(f"{path}: {e}")

def mergePresets(a, b):
    """
    Merge b into a. Values from b overwrite values from a.
    """
    for category in b:
        if category not in a:
            a[category] = {}
        for key, value in b[category].items():
            a[category][key] = value

def loadPresetChain(chain):
    """
    Given a list of preset names (or paths), load the whole chain.
    """
    assert len(chain) > 0

    preset = loadPreset(chain[0])
    for p in chain[1:]:
        newPreset = loadPreset(p)
        mergePresets(preset, newPreset)
    return preset

def validateSections(preset):
    """
    Perform a logic validation of the given preset - e.g., if a style is applied,
    validate all required keys are present. Ignores excessive keys.
    """
    VALID_SECTIONS = ["layout", "source", "tabs", "cuts", "framing", "tooling",
        "fiducials", "text", "text2", "text3", "text4", "page", "copperfill",
        "post", "debug"]
    extraSections = set(preset.keys()).difference(VALID_SECTIONS)
    if len(extraSections) != 0:
        raise PresetError(f"Extra sections {', '.join(extraSections)} in preset")
    missingSections = set(VALID_SECTIONS).difference(preset.keys())
    if len(missingSections) != 0:
        raise PresetError(f"Missing sections {', '.join(missingSections)} in preset")

def getPlacementClass(name):
    from kikit.panelize import (BasicGridPosition, OddEvenColumnPosition,
        OddEvenRowsPosition, OddEvenRowsColumnsPosition)
    mapping = {
        "none": BasicGridPosition,
        "rows": OddEvenRowsPosition,
        "cols": OddEvenColumnPosition,
        "rowsCols": OddEvenRowsColumnsPosition
    }
    try:
        return mapping[name]
    except KeyError:
        raise RuntimeError(f"Invalid alternation option '{name}' passed. " +
            "Valid options are: " + ", ".join(mapping.keys()))

def readSourceArea(specification, board):
    """
    Extract source area based on preset and a board.
    """
    try:
        type = specification["type"]
        tolerance = specification["tolerance"]
        if type == "auto":
            return expandRect(findBoardBoundingBox(board), tolerance)
        if type == "annotation":
            ref = specification["ref"]
            if ref.strip() == "":
                raise PresetError("When using source 'annotation' reference cannot be empty.")
            layer = specification.get("layer", Layer.Edge_Cuts)
            return expandRect(extractSourceAreaByAnnotation(board, ref, layer), tolerance)
        if type == "rectangle":
            tl = VECTOR2I(specification["tlx"], specification["tly"])
            br = VECTOR2I(specification["brx"], specification["bry"])
            return expandRect(BOX2I(tl, (br - tl)), tolerance)
        raise PresetError(f"Unknown type '{type}' of source specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'source'")

def setStackup(specification, panel):
    """
    Set stackup based on user preferences
    """
    stackup = specification["stack"]
    if stackup == "inherit":
        return
    if stackup == "2layer":
        panel.setCopperLayers(2)
    elif stackup == "4layer":
        panel.setCopperLayers(4)
    elif stackup == "6layer":
        panel.setCopperLayers(6)
    else:
        raise PresetError(f"Unknown type '{stackup}' of stack specification.")


def obtainPreset(presetPaths, validate=True, **kwargs):
    """
    Given a preset paths from the user and the overrides in the form of named
    arguments, construt the preset.

    Ensures a valid preset is always found
    """
    presetChain = [":default"] + list(presetPaths)
    preset = loadPresetChain(presetChain)
    for name, section in kwargs.items():
        if section is not None:
            mergePresets(preset, {name: section})
    if validate:
        validateSections(preset)
    postProcessPreset(preset)
    return preset

def buildLayout(preset, panel, sourceBoard, sourceArea):
    """
    Build layout for the boards - e.g., make a grid out of them.

    Return the list of created substrates and framing substrates. Also ensures
    that the partition line is build properly.
    """
    layout = preset["layout"]
    framing = preset["framing"]
    try:
        type = layout["type"]
        if type == "grid":
            placementClass = getPlacementClass(layout["alternation"])
            placer = placementClass(
                horSpace=layout["hspace"],
                verSpace=layout["vspace"],
                hbonewidth=layout["hbackbone"],
                vbonewidth=layout["vbackbone"],
                hboneskip=layout["hboneskip"],
                vboneskip=layout["vboneskip"],
                hbonefirst=layout["hbonefirst"],
                vbonefirst=layout["vbonefirst"],
                hevendiff=layout["hevendiff"],
                vevendiff=layout["vevendiff"])
            substrates = panel.makeGrid(
                boardfile=sourceBoard, sourceArea=sourceArea,
                rows=layout["rows"], cols=layout["cols"], destination=VECTOR2I(0, 0),
                rotation=layout["rotation"], placer=placer,
                netRenamePattern=layout["renamenet"], refRenamePattern=layout["renameref"],
                bakeText=layout["baketext"], bakeRef=layout["bakeref"])
            framingSubstrates = dummyFramingSubstrate(substrates, preset)
            panel.buildPartitionLineFromBB(framingSubstrates)
            backboneCuts = buildBackBone(layout, panel, substrates, framing)
            return substrates, framingSubstrates, backboneCuts
        if type == "plugin":
            lPlugin = layout["code"](preset, layout["arg"], layout["renamenet"],
                                     layout["renameref"], layout["vspace"],
                                     layout["hspace"], layout["rotation"])
            substrates = lPlugin.buildLayout(panel, sourceBoard, sourceArea)
            framingSubstrates = dummyFramingSubstrate(substrates, preset)
            lPlugin.buildPartitionLine(panel, framingSubstrates)
            backboneCuts = lPlugin.buildExtraCuts(panel)
            return substrates, framingSubstrates, backboneCuts

        raise PresetError(f"Unknown type '{type}' of layout specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'layout'")

def buildTabs(preset, panel, substrates, boundarySubstrates):
    """
    Build tabs for the substrates in between the boards. Return a list of cuts.
    """
    properties = preset["tabs"]
    try:
        type = properties["type"]
        if type == "none":
            return []
        if type == "fixed":
            panel.clearTabsAnnotations()
            panel.buildTabAnnotationsFixed(properties["hcount"],
                properties["vcount"], properties["hwidth"], properties["vwidth"],
                properties["mindistance"], boundarySubstrates)
            return panel.buildTabsFromAnnotations(properties["fillet"])
        if type == "spacing":
            panel.clearTabsAnnotations()
            panel.buildTabAnnotationsSpacing(properties["spacing"],
                properties["hwidth"], properties["vwidth"], boundarySubstrates)
            return panel.buildTabsFromAnnotations(properties["fillet"])
        if type == "corner":
            panel.clearTabsAnnotations()
            panel.buildTabAnnotationsCorners(properties["width"])
            return panel.buildTabsFromAnnotations(properties["fillet"])
        if type == "full":
            return panel.buildFullTabs(properties["cutout"], properties["patchcorners"])
        if type == "annotation":
            return panel.buildTabsFromAnnotations(properties["fillet"])
        if type == "plugin":
            pluginInst = properties["code"](preset, properties["arg"])
            return pluginInst.buildTabs(panel)
        raise PresetError(f"Unknown type '{type}' of tabs specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'tabs'")

def buildBackBone(layout, panel, substrates, frameSpace):
    """
    Append backbones to the panel. Return backbone cuts.
    """
    try:
        return panel.renderBackbone(layout["vbackbone"], layout["hbackbone"],
                                    layout["vbonecut"], layout["hbonecut"],
                                    layout["vboneskip"], layout["hboneskip"],
                                    layout["vbonefirst"], layout["hbonefirst"])
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'layout'")

def frameOffset(framing):
    try:
        type = framing["type"]
        if type == "none":
            return None, None
        if type == "railstb":
            return framing["vspace"], None
        if type == "railslr":
            return None, framing["hspace"]
        return framing["vspace"], framing["hspace"]
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'framing'")

def makeTabCuts(preset, panel, cuts):
    """
    Perform cuts on tab (does not ignore offset)
    """
    properties = preset["cuts"]
    if properties["type"] == "plugin":
        pluginInst = properties["code"](preset, properties["arg"])
        return pluginInst.renderTabCuts(panel, cuts)
    makeCuts(properties, panel, cuts, False)

def makeOtherCuts(preset, panel, cuts):
    """
    Perform non-tab cuts (ignore offset)
    """
    properties = preset["cuts"]
    if properties["type"] == "plugin":
        pluginInst = properties["code"](preset, properties["arg"])
        return pluginInst.renderOtherCuts(panel, cuts)
    makeCuts(properties, panel, cuts, True)

def makeCuts(properties, panel, cuts, ignoreOffset):
    """
    Perform cuts
    """
    try:
        type = properties["type"]
        offset = 0 if ignoreOffset else properties["offset"]
        if type == "none":
            return
        if type == "vcuts":
            panel.makeVCuts(cuts, properties["cutcurves"], offset)
            panel.vCutSettings.layer = properties["layer"]
            panel.vCutSettings.clearance = properties["clearance"]
            panel.vCutSettings.lineWidth = properties["linewidth"]
            panel.vCutSettings.textThickness = properties["textthickness"]
            panel.vCutSettings.textProlongation = properties["textprolongation"]
            panel.vCutSettings.endProlongation = properties["endprolongation"]
            panel.vCutSettings.textTemplate = properties["template"]
            panel.vCutSettings.textOffset = properties["textoffset"]

        elif type == "mousebites":
            panel.makeMouseBites(cuts, properties["drill"],
                properties["spacing"], offset, properties["prolong"])
        elif type == "layer":
            panel.makeCutsToLayer(cuts,
                layer=properties["layer"], prolongation=properties["prolong"],
                width=properties["linewidth"])
        else:
            raise PresetError(f"Unknown type '{type}' of cuts specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'cuts'")


def polygonToSubstrate(polygon):
    s = Substrate([])
    s.union(polygon)
    return s

def dummyFramingSubstrate(substrates, preset):
    """
    Generate dummy substrates that pretend to be the frame (to appear)
    """
    framingPreset = preset["framing"]
    if framingPreset["type"] == "plugin":
        pluginInst = framingPreset["code"](preset, framingPreset["arg"])
        return pluginInst.buildDummyFramingSubstrates(substrates)
    vSpace, hSpace = frameOffset(framingPreset)
    dummy = []
    minx, miny, maxx, maxy = substrates[0].bounds()
    for s in substrates:
        minx2, miny2, maxx2, maxy2 = s.bounds()
        minx = min(minx, minx2)
        miny = min(miny, miny2)
        maxx = max(maxx, maxx2)
        maxy = max(maxy, maxy2)
    # Note that the constructed substrates has to have a non-zero width/height.
    # If the width is zero, we break the input condition of the neighbor finding
    # algorithm (as there is no distinguishion between left and right side)
    width = fromMm(1)
    if vSpace is not None:
        top = box(minx, miny - 2 * vSpace - width, maxx, miny - 2 * vSpace)
        bottom = box(minx, maxy + 2 * vSpace, maxx, maxy + 2 * vSpace + width)
        dummy.append(polygonToSubstrate(top))
        dummy.append(polygonToSubstrate(bottom))
    if hSpace is not None:
        left = box(minx - 2 * hSpace - width, miny, minx - 2 * hSpace, maxy)
        right = box(maxx + 2 * hSpace, miny, maxx + 2 * hSpace + width, maxy)
        dummy.append(polygonToSubstrate(left))
        dummy.append(polygonToSubstrate(right))
    return dummy


def addFilletAndChamfer(preset, panel):
    """
    Add chamfer of frame based on the preset
    """
    chamferWidth, chamferHeight = preset["chamferwidth"], preset["chamferheight"]
    if chamferWidth < 0 or chamferHeight < 0:
        raise PresetError(f"Invalid chamfer value specified: [{chamferWidth}, {chamferHeight}]")
    fillet = preset["fillet"]
    if fillet < 0:
        raise PresetError(f"Invalid fillet value specified: {fillet}")
    if (chamferWidth != 0 or chamferHeight != 0) and fillet != 0:
        raise PresetError("You cannot specify both, chamfer and fillet. Set one of them to 0.")

    if fillet > 0:
        panel.addCornerFillets(fillet)
    if chamferWidth > 0 or chamferHeight > 0:
        panel.addCornerChamfers(chamferWidth, chamferHeight)


def buildFraming(preset, panel):
    """
    Build frame according to the preset and return cuts
    """
    framingPreset = preset["framing"]
    try:
        type = framingPreset["type"]
        if type == "none":
            return []
        if type == "railstb":
            panel.makeRailsTb(framingPreset["width"],
                framingPreset["mintotalheight"], framingPreset["maxtotalheight"],
                vspace=framingPreset["vspace"])
            addFilletAndChamfer(framingPreset, panel)
            return []
        if type == "railslr":
            panel.makeRailsLr(framingPreset["width"],
                framingPreset["mintotalwidth"], framingPreset["maxtotalwidth"],
                hspace=framingPreset["hspace"])
            addFilletAndChamfer(framingPreset, panel)
            return []
        if type == "frame":
            cuts = panel.makeFrame(framingPreset["width"], framingPreset["width"],
                framingPreset["hspace"], framingPreset["vspace"],
                framingPreset["mintotalwidth"], framingPreset["mintotalheight"],
                framingPreset["maxtotalwidth"], framingPreset["maxtotalheight"])
            addFilletAndChamfer(framingPreset, panel)
            if framingPreset["cuts"] == "both":
                return chain(*cuts)
            if framingPreset["cuts"] == "v":
                return cuts[0]
            if framingPreset["cuts"] == "h":
                return cuts[1]
            return []
        if type == "tightframe":
            panel.makeTightFrame(framingPreset["width"], framingPreset["width"],
                framingPreset["slotwidth"],  framingPreset["hspace"],
                framingPreset["vspace"], framingPreset["mintotalwidth"],
                framingPreset["mintotalheight"], framingPreset["maxtotalwidth"],
                framingPreset["maxtotalheight"])
            panel.boardSubstrate.removeIslands()
            addFilletAndChamfer(framingPreset, panel)
            return []
        if type == "plugin":
            framePlugin = framingPreset["code"](preset, framingPreset["arg"])
            return framePlugin.buildFraming(panel)
        raise PresetError(f"Unknown type '{type}' of frame specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'framing'")

def buildTabFillets(preset, panel, preFrameSubstrate):
    """
    Apply reverse tab fillets where tabs meet the support structure
    (backbones + frame). The support structure is computed from:
    - Frame geometry: diff of boardSubstrate before/after framing
    - Backbone geometry: stored during renderBackbone
    """
    # Local imports to avoid circular dependency (panelize_ui_impl <-> panelize)
    from kikit.panelize import addFrameFillets
    import shapely.ops

    fillet = preset["tabs"].get("fillet", 0)
    if fillet == 0:
        return

    supportParts = []
    frameGeometry = panel.boardSubstrate.substrates.difference(preFrameSubstrate)
    if not frameGeometry.is_empty:
        supportParts.append(frameGeometry)
    if panel.backbonePieces:
        supportParts.append(shapely.ops.unary_union(panel.backbonePieces))
    if not supportParts:
        return

    support = shapely.ops.unary_union(supportParts)
    filletedSupport = addFrameFillets(support, panel.substrates, fillet, panel)
    panel.boardSubstrate.union(filletedSupport)

def buildTooling(preset, panel):
    """
    Build tooling holes according to the preset
    """
    toolingPreset = preset["tooling"]
    try:
        type = toolingPreset["type"]
        if type == "none":
            return
        if type == "plugin":
            pluginInst = toolingPreset["code"](preset, toolingPreset["arg"])
            return pluginInst.buildTooling(panel)
        hoffset, voffset = toolingPreset["hoffset"], toolingPreset["voffset"]
        diameter = toolingPreset["size"]
        paste = toolingPreset["paste"]
        soldermaskmargin = toolingPreset["soldermaskmargin"]
        if type == "3hole":
            panel.addCornerTooling(3, hoffset, voffset, diameter, paste, soldermaskmargin)
            return
        if type == "4hole":
            panel.addCornerTooling(4, hoffset, voffset, diameter, paste, soldermaskmargin)
            return
        raise PresetError(f"Unknown type '{type}' of tooling specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'tooling'")

def buildFiducials(preset, panel):
    """
    Build tooling holes according to the preset
    """
    fidPreset = preset["fiducials"]
    try:
        type = fidPreset["type"]
        if type == "none":
            return
        if type == "plugin":
            pluginInst = fidPreset["code"](preset, fidPreset["arg"])
            return pluginInst.buildFiducials(panel)
        hoffset, voffset = fidPreset["hoffset"], fidPreset["voffset"]
        coppersize, opening = fidPreset["coppersize"], fidPreset["opening"]
        paste = fidPreset["paste"]
        if type == "3fid":
            panel.addCornerFiducials(3, hoffset, voffset, coppersize, opening, paste)
            return
        if type == "4fid":
            panel.addCornerFiducials(4, hoffset, voffset, coppersize, opening, paste)
            return
        raise PresetError(f"Unknown type '{type}' of fiducial specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'fiducials'")

def buildText(preset, panel):
    """
    Build text according to the preset
    """
    try:
        type = preset["type"]
        if type == "none":
            return
        # Since all boards are the same, we can use variables from the first project
        variables = kikitTextVars(panel.board, panel.projectVars[0])
        if preset["plugin"] is not None:
            variables.update(preset["plugin"](panel.board).variables())
        try:
            text = preset["text"].format(**variables)
        except KeyError as e:
            raise RuntimeError(f"Unknown variable {e} in text:\n{preset['text']}") from None
        if type == "simple":
            origin = resolveAnchor(preset["anchor"])(panel.boardSubstrate.boundingBox())
            origin += VECTOR2I(preset["hoffset"], preset["voffset"])

            panel.addText(
                text=text,
                position=origin,
                orientation=preset["orientation"],
                width=preset["width"],
                height=preset["height"],
                thickness=preset["thickness"],
                hJustify=preset["hjustify"],
                vJustify=preset["vjustify"],
                layer=preset["layer"])
            return
        raise PresetError(f"Unknown type '{type}' of text specification.")
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'text'")

def buildCopperfill(preset, panel):
    """
    Perform copperfill operation
    """
    try:
        type = preset["type"]
        if type == "none":
            return
        if type == "solid":
            panel.apply(SolidCopperFill(
                clearance=preset["clearance"],
                edgeclearance=preset["edgeclearance"],
                layers=preset["layers"],
            ))
        if type == "hatched":
            panel.apply(HatchedCopperFill(
                clearance=preset["clearance"],
                edgeclearance=preset["edgeclearance"],
                layers=preset["layers"],
                strokeWidth=preset["width"],
                strokeSpacing=preset["spacing"],
                orientation=preset["orientation"]
            ))
        if type == "hex":
            panel.apply(HexCopperFill(
                clearance=preset["clearance"],
                edgeclearance=preset["edgeclearance"],
                layers=preset["layers"],
                diameter=preset["diameter"],
                space=preset["spacing"],
                threshold=preset["threshold"]
            ))
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'postprocessing'")

def buildPostprocessing(preset, panel):
    """
    Perform postprocessing operations
    """
    try:
        type = preset["type"]
        if type != "auto":
            raise PresetError(f"Unknown type '{type}' of postprocessing specification.")
        if preset["millradius"] > 0 and preset["millradiusouter"] > 0:
            raise PresetError("You cannot specify both, millradius and millradiusouter")
        if preset["millradius"] > 0:
            panel.addMillFillets(preset["millradius"])
        if preset["millradiusouter"] > 0:
            panel.addTabMillFillets(preset["millradiusouter"])
        if preset["copperfill"]:
            panel.copperFillNonBoardAreas()
        if preset["origin"]:
            bBox = panel.boardSubstrate.boundingBox()
            origin = resolveAnchor(preset["origin"])(bBox)
            panel.setAuxiliaryOrigin(origin)
            panel.setGridOrigin(origin)
        if preset["dimensions"]:
            panel.addPanelDimensions(Layer.Dwgs_User, fromMm(5))
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'postprocessing'")

def runUserScript(preset, panel):
    """
    Run post processing script
    """
    if preset["script"] == "":
        return
    import importlib.util

    spec = importlib.util.spec_from_file_location("kikit.user_script",
        preset["script"])
    userScriptModule = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(userScriptModule)
    userScriptModule.kikitPostprocess(panel, preset["scriptarg"])


def buildDebugAnnotation(preset, panel):
    """
    Add debug annotation to the panel
    """
    try:
        if preset["drawPartitionLines"]:
            panel.debugRenderPartitionLines()
        if preset["drawBackboneLines"]:
            panel.debugRenderBackboneLines()
        if preset["drawboxes"]:
            panel.debugRenderBoundingBoxes()
        if preset["drawTabFillet"]:
            panel.debugRenderTabFillet()
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'debug'")


def positionPanel(preset, panel):
    """
    Position the panel on the paper
    """
    try:
        bBox = panel.boardSubstrate.boundingBox()
        pageSize = panel.getPageDimensions()
        posx = preset["posx"] * pageSize[0] if isinstance(preset["posx"], PercentageValue) else preset["posx"]
        posy = preset["posy"] * pageSize[1] if isinstance(preset["posy"], PercentageValue) else preset["posy"]
        origin = resolveAnchor(preset["anchor"])(bBox)
        translateVec = (-origin[0] + posx, -origin[1] + posy)
        panel.translate(translateVec)
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'page'")

def setPageSize(preset, panel, sourceBoard):
    """
    Set page size of the panel file
    """
    try:
        pageSize = preset["type"]
        if pageSize == "inherit":
            panel.inheritPageSize(sourceBoard)
            return
        if pageSize == "user":
            panel.setPageSize((preset["width"], preset["height"]))
            return
        panel.setPageSize(pageSize)
    except KeyError as e:
        raise PresetError(f"Missing parameter '{e}' in section 'page'")

HookPluginInvoker = Callable[[HookPlugin], None]

def loadHookPlugins(pluginSpec: List[Tuple[str, str, str]], board: pcbnew.BOARD,
                   preset: Dict[str, Dict[str, Any]]) -> Callable[[HookPluginInvoker], None]:
    """
    Loads hook plugins based on the specification and returns a function that
    will invoke given callable on each of the loaded plugins.

    This function assumes it is called only once during the whole execution of
    the process.
    """
    plugins: List[HookPlugin] = []
    for moduleName, pluginName, arg in pluginSpec:
        try:
            if moduleName.endswith(".py"):
                plugin = loadHookPluginFromFile(moduleName, pluginName, arg, board, preset, len(plugins))
                plugins.append(plugin)
            else:
                plugin = loadHookPluginFromModule(moduleName, pluginName, arg, board, preset)
                plugins.append(plugin)
        except Exception as e:
            raise RuntimeError(f"Cannot instantiate '{moduleName}:{pluginName}': {e}") from None

    def usePlugins(invoker: HookPluginInvoker) -> None:
        nonlocal plugins
        for p in plugins:
            invoker(p)
    return usePlugins

def loadHookPluginFromFile(moduleName: str, pluginName: str, arg: str,
                           board: pcbnew.BOARD, preset: Dict[str, Dict[str, Any]],
                           seq: int) -> HookPlugin:
    import importlib.util

    if not os.path.exists(moduleName):
        raise RuntimeError(f"File doesn't exist")

    spec = importlib.util.spec_from_file_location(
                f"kikit.user.hook_plugin{seq}",
                moduleName)
    if spec is None:
        raise RuntimeError(f"Plugin module '{moduleName}' doesn't exist")
    pluginModule = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(pluginModule)
    pluginType = getattr(pluginModule, pluginName)
    return pluginType(arg, board, preset)

def loadHookPluginFromModule(moduleName: str, pluginName: str, arg: str,
                           board: pcbnew.BOARD, preset: Dict[str, Dict[str, Any]]) \
                                -> HookPlugin:
    import importlib
    pluginModule = importlib.import_module(moduleName)
    pluginType = getattr(pluginModule, pluginName)
    return pluginType(arg, board, preset)



from dataclasses import dataclass
import os
from typing import Any, List
from kikit import plugin
from kikit.units import readLength, readAngle, readPercents
from kikit.defs import Layer, EDA_TEXT_HJUSTIFY_T, EDA_TEXT_VJUSTIFY_T, PAPER_SIZES

class PresetError(RuntimeError):
    pass

ANCHORS = ["tl", "tr", "bl", "br", "mt", "mb", "ml", "mr", "c"]
PAPERS = ["inherit"] + PAPER_SIZES + ["user"]

@dataclass
class FootprintId:
    lib: str
    footprint: str

class SectionBase:
    def __init__(self, isGuiRelevant, description):
        self.description = description
        self.isGuiRelevant = isGuiRelevant

    def validate(self, x: str) -> Any:
        raise NotImplementedError("Validate was not overridden for SectionBase")

class SLength(SectionBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self, x):
        return readLength(x)

class SPercent(SectionBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self, x):
        x = x.strip()
        if not x.endswith("%"):
            raise PresetError("Percentage error has to end with %")
        return readPercents(x)

class SLengthOrPercent(SectionBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self, x):
        x = x.strip()
        if x.endswith("%"):
            return readPercents(x)
        return readLength(x)

class SAngle(SectionBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self, x):
        return readAngle(x)

class SNum(SectionBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self, x):
        return int(x)

class SNaturalNum(SNum):
    def validate(self, x):
        val = int(x)
        if val < 0:
            raise PresetError(f"A non-negative number expected, got '{x}'")
        return val

class SStr(SectionBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self, x):
        return str(x)

class SPlugin(SectionBase):
    seq: int = 0

    def __init__(self, pluginType, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pluginType = pluginType

    def validate(self, x):
        if x == "none":
            return None
        self.seq += 1

        pieces = str(x).rsplit(".", maxsplit=1)
        if len(pieces) != 2:
            raise RuntimeError(f"Invalid plugin specification '{x}'")
        moduleName, pluginName = pieces[0], pieces[1]
        plugin = self.loadFromFile(moduleName, pluginName) if moduleName.endswith(".py") \
                 else self.loadFromModule(moduleName, pluginName)
        if not issubclass(plugin, self.pluginType):
            raise RuntimeError(f"Invalid plugin type specified, {self.pluginType.__name__} expected")
        setattr(plugin, "__kikit_preset_repr", x)
        return plugin

    def loadFromFile(self, file, name):
        import importlib.util

        if not os.path.exists(file):
            raise RuntimeError(f"File {file} doesn't exist")
        spec = importlib.util.spec_from_file_location(
                f"kikit.user.SPlugin_{self.seq}",
                file)
        if spec is None:
            raise RuntimeError(f"Plugin module '{file}' doesn't exist")
        pluginModule = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(pluginModule)
        return getattr(pluginModule, name)

    def loadFromModule(self, module, name):
        import importlib
        pluginModule = importlib.import_module(module)
        return getattr(pluginModule, name)

class SChoiceBase(SectionBase):
    def __init__(self, vals, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vals = vals

    def validate(self, s):
        if s not in self.vals:
            c = ", ".join(self.vals)
            raise PresetError(f"'{s}' is not allowed Use one of {c}.")
        return s

class SChoice(SChoiceBase):
    def __init__(self, vals, *args, **kwargs):
        super().__init__(vals, *args, **kwargs)

class SBool(SChoiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(["True", "False"], *args, **kwargs)

    def validate(self, s):
        if isinstance(s, bool):
            return s
        if isinstance(s, str):
            sl = str(s).lower()
            if sl in ["1", "true", "yes"]:
                return True
            if sl in ["0", "false", "no"]:
                return False
            raise PresetError(f"Uknown boolean value '{s}'")
        raise PresetError(f"Got {s}, expected boolean value")

class SHJustify(SChoiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(["left", "right", "center"], *args, **kwargs)

    def validate(self, s):
        choices = {
            "left": EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_LEFT,
            "right": EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_RIGHT,
            "center": EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_CENTER
        }
        if s in choices:
            return choices[s]
        raise PresetError(f"'{s}' is not valid justification value")

class SVJustify(SChoiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(["top", "bottom", "center"], *args, **kwargs)

    def validate(self, s):
        choices = {
            "top": EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_TOP,
            "center": EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_CENTER,
            "bottom": EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_BOTTOM
        }
        if s in choices:
            return choices[s]
        raise PresetError(f"'{s}' is not valid justification value")

class SLayer(SChoiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(
            [item.name.replace("Layer.", "").replace("_", ".")
                for item in Layer], *args, **kwargs)

    def validate(self, s):
        if isinstance(s, int) or s.isdigit():
            if int(s) in tuple(item.value for item in Layer):
                return Layer(int(s))
            raise PresetError(f"{s} is not a valid layer number")
        if isinstance(s, str):
            try:
                return Layer[s.replace(".", "_")]
            except Exception:
                pass
        raise PresetError(f"Got {s}, expected layer name or number")

class SList(SectionBase):
    def validate(self, x: str) -> Any:
        return [v.strip() for v in x.split(",")]

class SLayerList(SList):
    def __init__(self, isGuiRelevant, description, shortcuts={}):
        super().__init__(isGuiRelevant, description)
        self._shortcuts = shortcuts

    def validate(self, x: str) -> Any:
        if x in self._shortcuts:
            return self._shortcuts[x]
        return [self.readLayer(x) for x in super().validate(x)]

    def readLayer(self, s: str) -> Layer:
        if isinstance(s, int) or s.isdigit():
            if int(s) in tuple(item.value for item in Layer):
                return Layer(int(s))
            raise PresetError(f"{s} is not a valid layer number")
        if isinstance(s, str):
            try:
                return Layer[s.replace(".", "_")]
            except Exception:
                pass
        raise PresetError(f"Got {s}, expected layer name or number")

class SFootprintList(SList):
    def validate(self, x: str) -> Any:
        result: List[FootprintId] = []
        for v in super().validate(x):
            s = v.split(":", 1)
            if len(s) != 2:
                PresetError(f"'{v}' is not a valid footprint name in the form '<lib>:<footprint>'")
            result.append(FootprintId(s[0], s[1]))
        return result


def validateSection(name, sectionDefinition, section):
    try:
        for key, validator in sectionDefinition.items():
            if key not in section:
                continue
            section[key] = validator.validate(section[key])
    except Exception as e:
        raise PresetError(f"Error in section {name}: {e}")
    return section

def typeIn(values):
    return lambda section: section["type"] in values

def always():
    return lambda section: True

def never():
    return lambda section: False

LAYOUT_SECTION = {
    "type": SChoice(
        ["grid", "plugin"],
        always(),
        "Layout type"),
    "alternation": SChoice(
        ["none", "rows", "cols", "rowsCols"],
        typeIn(["grid", "plugin"]),
        "Specify alternations of board rotation"),
    "hspace": SLength(
        always(),
        "Specify horizontal gap between the boards"),
    "vspace": SLength(
        always(),
        "Specify vertical gap between the boards"),
    "space": SLength(
        never(),
        "Specify the gap between the boards in both direction"),
    "hevendiff": SLength(
        always(),
        "Extra spacing added to even spaces between boards columns"),
    "vevendiff": SLength(
        always(),
        "Extra spacing added to even spaces between boards rows"),
    "hbackbone": SLength(
        typeIn(["grid", "plugin"]),
        "The width of horizontal backbone (0 means no backbone)"),
    "vbackbone": SLength(
        typeIn(["grid", "plugin"]),
        "The width of vertical backbone (0 means no backbone)"),
    "hboneskip": SNaturalNum(
        typeIn(["grid", "plugin"]),
        "Skip every given number of horizontal backbones"),
    "vboneskip": SNaturalNum(
        typeIn(["grid", "plugin"]),
        "Skip every given number of vertical backbones"),
    "hbonefirst": SNaturalNum(
        typeIn(["grid", "plugin"]),
        "Specify first horizontal backbone to render"),
    "vbonefirst": SNaturalNum(
        typeIn(["grid", "plugin"]),
        "Specify first vertical backbone to render"),
    "rotation": SAngle(
        always(),
        "Rotate the boards before placing them in the panel"),
    "rows": SNaturalNum(
        typeIn(["grid", "plugin"]),
        "Specify the number of rows in the grid pattern"),
    "cols": SNaturalNum(
        typeIn(["grid", "plugin"]),
        "Specify the number of columns in the grid pattern"),
    "vbonecut": SBool(
        typeIn(["grid", "plugin"]),
        "Add cuts in vertical backbones for easier depanelization"),
    "hbonecut": SBool(
        typeIn(["grid", "plugin"]),
        "Add cuts in horizontal backbones for easier depanelization"),
    "renamenet": SStr(
        always(),
        "Net renaming pattern"),
    "renameref": SStr(
        always(),
        "Reference renaming pattern"),
    "baketext": SBool(
        always(),
        "Substitute variables in text elements"
    ),
    "bakeref": SBool(
        always(),
        "Bake old references before renaming"
    ),
    "code": SPlugin(
        plugin.LayoutPlugin,
        typeIn(["plugin"]),
        "Plugin specification as moduleName.pluginName"),
    "arg": SStr(
        typeIn(["plugin"]),
        "String argument for the layout plugin")
}

def ppLayout(section):
    section = validateSection("layout", LAYOUT_SECTION, section)
    # The space parameter overrides hspace and vspace
    if "space" in section:
        section["hspace"] = section["vspace"] = section["space"]

SOURCE_SECTION = {
    "type": SChoice(
        ["auto", "rectangle", "annotation"],
        always(),
        "Source type"),
    "tolerance": SLength(
        typeIn(["auto", "annotation"]),
        "Tolerance enlarges the source area by given amount"),
    "tlx": SLength(
        typeIn(["rectangle"]),
        "Corner of the rectangle"),
    "tly": SLength(
        typeIn(["rectangle"]),
        "Corner of the rectangle"),
    "brx": SLength(
        typeIn(["rectangle"]),
        "Corner of the rectangle"),
    "bry": SLength(
        typeIn(["rectangle"]),
        "Corner of the rectangle"),
    "ref": SStr(
        typeIn(["annotation"]),
        "Specify reference of KiKit annotation symbol"),
    "layer": SLayer(
        typeIn(["annotation"]),
        "Specify layer for annotation lines (default: Edge.Cuts)"),
    "stack": SChoice(
        ["inherit", "2layer", "4layer", "6layer"],
        always(),
        "Specify the number of layers of the panel")
}

def ppSource(section):
    section = validateSection("source", SOURCE_SECTION, section)

TABS_SECTION = {
    "type": SChoice(
        ["none", "fixed", "spacing", "full", "corner", "annotation", "plugin"],
        always(),
        "Tab type"),
    "vwidth": SLength(
        typeIn(["fixed", "spacing", "plugin"]),
        "Specify width of vertical tabs"),
    "hwidth": SLength(
        typeIn(["fixed", "spacing", "plugin"]),
        "Specify width of horizontal tabs"),
    "width": SLength(
        typeIn(["corner", "plugin"]),
        "Specify tab width"),
    "mindistance": SLength(
        typeIn(["fixed", "plugin"]),
        "Minimal spacing between the tabs. If there are too many tabs, their count is reduced."),
    "spacing": SLength(
        typeIn(["spacing", "plugin"]),
        "The maximum spacing of the tabs."),
    "vcount": SNum(
        typeIn(["fixed", "plugin"]),
        "Number of tabs in a given direction."),
    "hcount": SNum(
        typeIn(["fixed", "plugin"]),
        "Number of tabs in a given direction."),
    "cutout": SLength(
        typeIn(["fixed", "full", "plugin"]),
        "Depth of cutouts into the frame"),
    "patchcorners": SBool(
        typeIn(["fixed", "full", "plugin"]),
        "Choose if to apply corner patches for the full tabs"
    ),
    "tabfootprints": SFootprintList(
        typeIn(["annotation", "plugin"]),
        "Specify custom footprints that will be used for tab annotations."),
    "fillet": SLength(
        typeIn(["fixed", "spacing", "corner", "annotation", "plugin"]),
        "Specify tab fillet radius (experimental)"
    ),
    "code": SPlugin(
        plugin.TabsPlugin,
        typeIn(["plugin"]),
        "Plugin specification as moduleName.pluginName"),
    "arg": SStr(
        typeIn(["plugin"]),
        "String argument for the layout plugin")
}

def ppTabs(section):
    section = validateSection("tabs", TABS_SECTION, section)
    if "width" in section:
        section["vwidth"] = section["hwidth"] = section["width"]

CUTS_SECTION = {
    "type": SChoice(
        ["none", "mousebites", "vcuts", "layer", "plugin"],
        always(),
        "Cut type"),
    "drill": SLength(
        typeIn(["mousebites", "plugin"]),
        "Drill diameter"),
    "spacing": SLength(
        typeIn(["mousebites", "plugin"]),
        "Hole spacing"),
    "offset": SLength(
        typeIn(["mousebites", "vcuts", "plugin"]),
        "Offset cuts into the board"),
    "prolong": SLength(
        typeIn(["mousebites", "layer", "plugin"]),
        "Tangentiall prolong cuts (to cut mill fillets)"),
    "clearance": SLength(
        typeIn(["vcuts", "plugin"]),
        "Add copper clearance around V-cuts"),
    "cutcurves": SBool(
        typeIn(["vcuts", "plugin"]),
        "Approximate curves with straight cut"),
    "linewidth": SLength(
        typeIn(["vcuts", "layer", "plugin"]),
        "Line width to plot cuts with"),
    "textthickness": SLength(
        typeIn(["vcuts", "plugin"]),
        "Text thickness for width"),
    "textsize": SLength(
        typeIn(["vcuts", "plugin"]),
        "Text size for vcuts"),
    "endprolongation": SLength(
        typeIn(["vcuts", "plugin"]),
        "Prolongation on the end of V-CUT without text"),
    "textprolongation": SLength(
        typeIn(["vcuts", "plugin"]),
        "Prolongation of the text size of V-CUT"),
    "textoffset": SLength(
        typeIn(["vcuts", "plugin"]),
        "Text offset from the V-CUT"),
    "template": SStr(
        typeIn(["vcuts", "plugin"]),
        "Text template for the V-CUT"),
    "layer": SLayer(
        typeIn(["vcuts", "layer", "plugin"]),
        "Specify layer for the drawings"),
    "code": SPlugin(
        plugin.CutsPlugin,
        typeIn(["plugin"]),
        "Plugin specification as moduleName.pluginName"),
    "arg": SStr(
        typeIn(["plugin"]),
        "String argument for the layout plugin")
}

def ppCuts(section):
    section = validateSection("cuts", CUTS_SECTION, section)

FRAMING_SECTION = {
    "type": SChoice(
        ["none", "railstb", "railslr", "frame", "tightframe", "plugin"],
        always(),
        "Framing type"),
    "hspace": SLength(
        typeIn(["frame", "railslr", "tightframe", "plugin"]),
        "Horizontal space between PCBs and the frame"),
    "vspace": SLength(
        typeIn(["frame", "railstb", "tightframe", "plugin"]),
        "Vertical space between PCBs and the frame"),
    "space": SLength(
        never(),
        "Space between frame/rails and PCBs"),
    "width": SLength(
        typeIn(["frame", "railstb", "railslr", "tightframe", "plugin"]),
        "Width of the framing"),
    "mintotalheight": SLength(
        typeIn(["frame", "railstb", "tightframe", "plugin"]),
        "Minimal height of the panel"
    ),
    "mintotalwidth": SLength(
        typeIn(["frame", "raillr", "tightframe", "plugin"]),
        "Minimal width of the panel"
    ),
    "maxtotalheight": SLength(
        typeIn(["frame", "railstb", "tightframe", "plugin"]),
        "Maximal height of the panel"
    ),
    "maxtotalwidth": SLength(
        typeIn(["frame", "raillr", "tightframe", "plugin"]),
        "Maximal width of the panel"
    ),
    "slotwidth": SLength(
        typeIn(["tightframe", "plugin"]),
        "Width of the milled slot"),
    "cuts": SChoice(
        ["none", "both", "v", "h"],
        typeIn(["frame", "plugin"]),
        "Add cuts to the corners of the frame"),
    "chamferwidth": SLength(
        typeIn(["tightframe", "frame", "railslr", "railstb", "plugin"]),
        "Add chamfer to the 4 corners of the panel. Specify chamfer width."),
    "chamferheight": SLength(
        typeIn(["tightframe", "frame", "railslr", "railstb", "plugin"]),
        "Add chamfer to the 4 corners of the panel. Specify chamfer height."),
    "chamfer": SLength(
        never(),
        "Add chamfer to the 4 corners of the panel. Specifies a 45° chamfer."),
    "fillet": SLength(
        typeIn(["tightframe", "frame", "railslr", "railstb", "plugin"]),
        "Add fillet to the 4 corners of the panel. Specify fillet radius."),
    "code": SPlugin(
        plugin.FramingPlugin,
        typeIn(["plugin"]),
        "Plugin specification as moduleName.pluginName"),
    "arg": SStr(
        typeIn(["plugin"]),
        "String argument for the layout plugin")
}

def ppFraming(section):
    section = validateSection("framing", FRAMING_SECTION, section)
    # The space parameter overrides hspace and vspace
    if "space" in section:
        section["hspace"] = section["vspace"] = section["space"]
    if "chamfer" in section:
        section["chamferwidth"] = section["chamferheight"] = section["chamfer"]

TOOLING_SECTION = {
    "type": SChoice(
        ["none", "3hole", "4hole", "plugin"],
        always(),
        "Tooling type"),
    "hoffset": SLength(
        typeIn(["3hole", "4hole", "plugin"]),
        "Horizontal offset for the hole"),
    "voffset": SLength(
        typeIn(["3hole", "4hole", "plugin"]),
        "Vertical offset for the hole"),
    "size": SLength(
        typeIn(["3hole", "4hole", "plugin"]),
        "Hole diameter"),
    "paste": SBool(
        typeIn(["3hole", "4hole", "plugin"]),
        "Include holes on the paste layer"),
    "soldermaskmargin": SLength(
        typeIn(["3hole", "4hole", "plugin"]),
        "Solder mask expansion/margin"),
    "code": SPlugin(
        plugin.ToolingPlugin,
        typeIn(["plugin"]),
        "Plugin specification as moduleName.pluginName"),
    "arg": SStr(
        typeIn(["plugin"]),
        "String argument for the layout plugin")
}

def ppTooling(section):
    section = validateSection("tooling", TOOLING_SECTION, section)

FIDUCIALS_SECTION = {
    "type": SChoice(
        ["none", "3fid", "4fid", "plugin"],
        always(),
        "Fiducial type"),
    "hoffset": SLength(
        typeIn(["3fid", "4fid", "plugin"]),
        "Horizontal offset for the fiducial"),
    "voffset": SLength(
        typeIn(["3fid", "4fid", "plugin"]),
        "Vertical offset for the fiducial"),
    "coppersize": SLength(
        typeIn(["3fid", "4fid", "plugin"]),
        "Diameter of the copper part"),
    "opening": SLength(
        typeIn(["3fid", "4fid", "plugin"]),
        "Diameter of the opening"),
    "paste": SBool(
        typeIn(["3fid", "4fid", "plugin"]),
        "Include fiducials on the paste layer"),
    "code": SPlugin(
        plugin.FiducialsPlugin,
        typeIn(["plugin"]),
        "Plugin specification as moduleName.pluginName"),
    "arg": SStr(
        typeIn(["plugin"]),
        "String argument for the layout plugin")
}

def ppFiducials(section):
    section = validateSection("fiducials", FIDUCIALS_SECTION, section)

TEXT_SECTION = {
    "type": SChoice(
        ["none", "simple"],
        always(),
        "Text type"),
    "hoffset": SLength(
        typeIn(["simple"]),
        "Horizontal offset of the text from anchor"),
    "voffset": SLength(
        typeIn(["simple"]),
        "Vertical offset of the text from anchor"),
    "width": SLength(
        typeIn(["simple"]),
        "Width of a character"),
    "height": SLength(
        typeIn(["simple"]),
        "Height of a character"),
    "thickness": SLength(
        typeIn(["simple"]),
        "Thickness of a character"),
    "hjustify": SHJustify(
        typeIn(["simple"]),
        "Text alignment"),
    "vjustify": SVJustify(
        typeIn(["simple"]),
        "Text alignment"),
    "layer": SLayer(
        typeIn(["simple"]),
        "Text layer"),
    "orientation": SAngle(
        typeIn(["simple"]),
        "Orientation of the text"),
    "text": SStr(
        typeIn(["simple"]),
        "Text to render"),
    "anchor": SChoice(
        ANCHORS,
        typeIn(["simple"]),
        "Anchor for positioning the text"),
    "plugin": SPlugin(
        plugin.TextVariablePlugin,
        typeIn(["simple"]),
        "Plugin for extra text variables")
}

def ppText(section):
    section = validateSection("text", TEXT_SECTION, section)

COPPERFILL_SECTION = {
    "type": SChoice(
        ["none", "solid", "hatched", "hex"],
        always(),
        "Fill non board areas with copper"),
    "clearance": SLength(
        typeIn(["solid", "hatched", "hex"]),
        "Clearance between the fill and boards"),
    "edgeclearance": SLength(
        typeIn(["solid", "hatched", "hex"]),
        "Clearance between the fill and boards"),
    "layers": SLayerList(
        typeIn(["solid", "hatched", "hex"]),
        "Specify which layer to fill with copper",
        {
            "all": Layer.allCu()
        }),
    "width": SLength(
        typeIn(["hatched"]),
        "Width of hatch strokes"),
    "spacing": SLength(
        typeIn(["hatched", "hex"]),
        "Spacing of hatch strokes or hexagons"),
    "orientation": SAngle(
        typeIn(["hatched"]),
        "Orientation of the strokes"),
    "diameter": SLength(
        typeIn(["hex"]),
        "Diameter of hexagons"
    ),
    "threshold": SPercent(
        typeIn(["hex"]),
        "Remove fragments smaller than threshold"
    )
}

def ppCopper(section):
    section = validateSection("copperfill", COPPERFILL_SECTION, section)

POST_SECTION = {
    "type": SChoice(
        ["auto"],
        never(),
        "Postprocessing type"),
    "copperfill": SBool(
        always(),
        "DEPRECATED, use section copperfill instead. Fill unused areas of the panel with copper"),
    "millradius": SLength(
        always(),
        "Simulate milling operation"),
    "millradiusouter": SLength(
        always(),
        "Simulate milling operation only on the outer perimeter of boards"),
    "reconstructarcs": SBool(
        always(),
        "Try to reconstruct arcs"),
    "refillzones": SBool(
        always(),
        "Refill all zones in the panel"),
    "script": SStr(
        always(),
        "Specify path to a custom postprocessing script"),
    "scriptarg": SStr(
        always(),
        "String argument for the postprocessing script"),
    "origin": SChoice(
        ANCHORS + [""],
        always(),
        "Place auxiliary origin"),
    "dimensions": SBool(
        always(),
        "Add dimensions markings to the finished panel"
    ),
    "edgewidth": SLength(
        always(),
        "Specify line width for the Edge.Cuts of the panel"
    )
}

def ppPost(section):
    section = validateSection("post", POST_SECTION, section)

PAGE_SECTION = {
    "type": SChoice(
        PAPERS,
        always(),
        "Size of paper"),
    "anchor": SChoice(
        ANCHORS,
        always(),
        "Anchor for positioning the panel on the page"),
    "posx": SLengthOrPercent(
        always(),
        "X position of the panel. Length or percents of page width."),
    "posy": SLengthOrPercent(
        always(),
        "Y position of the panel. Length or percents of page height."),
    "width": SLength(
        typeIn(["user"]),
        "Width of custom paper"),
    "height": SLength(
        typeIn(["user"]),
        "Height of custom paper"),
}

def ppPage(section):
    section = validateSection("page", PAGE_SECTION, section)

DEBUG_SECTION = {
    "type": SChoice(
        ["none"],
        never(),
        ""),
    "drawPartitionLines": SBool(
        always(),
        "Draw parition lines"),
    "drawBackboneLines": SBool(
        always(),
        "Draw backbone lines"),
    "drawboxes": SBool(
        always(),
        "Draw board bounding boxes"),
    "trace": SBool(
        always(),
        "Print stacktrace"),
    "drawtabfail": SBool(
        always(),
        "Visualize tab building failures"
    ),
    "deterministic": SBool(
        always(),
        "Make KiCAD IDs deterministic"),
    "drawTabFillet": SBool(
        always(),
        "Draw forward tabs, reverse tabs, and raw frame geometry for fillet debugging")
}

def ppDebug(section):
    section = validateSection("debug", DEBUG_SECTION, section)

availableSections = {
    "Layout": LAYOUT_SECTION,
    "Source": SOURCE_SECTION,
    "Tabs": TABS_SECTION,
    "Cuts": CUTS_SECTION,
    "Framing": FRAMING_SECTION,
    "Tooling": TOOLING_SECTION,
    "Fiducials": FIDUCIALS_SECTION,
    "Text": TEXT_SECTION,
    "Text2": TEXT_SECTION,
    "Text3": TEXT_SECTION,
    "Text4": TEXT_SECTION,
    "Copperfill": COPPERFILL_SECTION,
    "Page": PAGE_SECTION,
    "Post": POST_SECTION,
    "Debug": DEBUG_SECTION,
}
